# `apis` sdk is used to prgrammatically call the algorithms.

import time

from cfcreator import *
from PIL import Image
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
from pydantic import BaseModel
from collections import OrderedDict
from cftool.misc import get_err_msg
from cftool.misc import random_hash
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict
from cftool.data_structures import Workflow
from cftool.data_structures import WorkNode
from cftool.data_structures import InjectionPack
from cfclient.core import HttpClient
from cfclient.utils import download_image_with_retry
from cfclient.models import TextModel
from cfclient.models import ImageModel
from cfclient.models import algorithms as registered_algorithms
from cflearn.api.cv.diffusion import ControlNetHints


TRes = Union[List[Union[str, int, float]], List[Image.Image]]
UPLOAD_ENDPOINT = "$upload"
ADD_TEXT_ENDPOINT = "$add_text"
CONTROL_HINT_ENDPOINT = "$control_hint"
FOR_EACH_ENDPOINT = "$for_each"
PICK_ENDPOINT = "$pick"
ALL_LATENCIES_KEY = "$all_latencies"
EXCEPTION_MESSAGE_KEY = "$exception_message"
endpoint2method = {
    txt2img_sd_endpoint: "txt2img",
    txt2img_sd_inpainting_endpoint: "sd_inpainting",
    txt2img_sd_outpainting_endpoint: "sd_outpainting",
    img2img_sd_endpoint: "img2img",
    img2img_sr_endpoint: "sr",
    img2img_sod_endpoint: "sod",
    img2img_inpainting_endpoint: "inpainting",
    img2img_semantic2img_endpoint: "semantic2img",
    img2img_harmonization_endpoint: "harmonization",
    img2txt_caption_endpoint: "image_captioning",
    txt2txt_prompt_enhance_endpoint: "prompt_enhance",
    new_control_multi_endpoint: "run_multi_controlnet",
    upscale_tile_endpoint: "upscale_tile",
    cv_blur_endpoint: "blur",
    cv_grayscale_endpoint: "grayscale",
    cv_erode_endpoint: "erode",
    cv_resize_endpoint: "resize",
    cv_affine_endpoint: "affine",
    cv_get_mask_endpoint: "get_mask",
    cv_inverse_endpoint: "inverse",
    cv_fill_bg_endpoint: "fill_bg",
    cv_get_size_endpoint: "get_size",
    cv_modify_box_endpoint: "modify_box",
    cv_generate_masks_endpoint: "generate_masks",
    cv_crop_image_endpoint: "crop_image",
    cv_histogram_match_endpoint: "histogram_match",
    cv_image_similarity_endpoint: "image_similarity",
    cv_repositioning_endpoint: "repositioning",
    facexlib_parse_endpoint: "facexlib_parse",
    facexlib_detect_endpoint: "facexlib_detect",
    paste_pipeline_endpoint: "paste_pipeline",
    UPLOAD_ENDPOINT: "get_image",
    ADD_TEXT_ENDPOINT: "add_text",
    CONTROL_HINT_ENDPOINT: "get_control_hint",
    FOR_EACH_ENDPOINT: "for_each",
    PICK_ENDPOINT: "pick",
}


class ForEachModel(BaseModel):
    endpoint: str
    loops: Dict[str, List[Any]]
    params: Dict[str, Any]
    loop_backs: Optional[List[InjectionPack]] = None


class PickModel(BaseModel):
    index: int
    values: List[Any]


class APIs:
    algorithms: Dict[str, IAlgorithm]

    def __init__(
        self,
        *,
        clients: Optional[dict] = None,
        algorithms: Optional[Dict[str, IAlgorithm]] = None,
        focuses_endpoints: Optional[List[str]] = None,
        excludes_endpoints: Optional[List[str]] = None,
        verbose: Optional[Union[str, bool]] = "auto",
        lazy_load: Optional[bool] = True,
    ) -> None:
        if focuses_endpoints is None:
            focuses = None
        else:
            focuses = list(map(endpoint2algorithm, focuses_endpoints))
        if excludes_endpoints is None:
            excludes = None
        else:
            excludes = list(map(endpoint2algorithm, excludes_endpoints))
        if verbose is not None:
            if verbose == "auto":
                verbose = focuses is not None
            OPT["verbose"] = verbose
        if lazy_load is not None:
            OPT["lazy_load"] = lazy_load

        self._is_new_http_client = False
        if clients is None:
            self._is_new_http_client = True
            self._http_client = HttpClient()
            clients = dict(http=self._http_client, triton=None)
        else:
            self._http_client = clients.get("http")
            if self._http_client is None:
                print_warning(f"cannot find `http` client in {clients}, creating one")
                self._is_new_http_client = True
                self._http_client = HttpClient()
                clients["http"] = self._http_client
        if algorithms is not None:
            self.algorithms = algorithms
        else:
            self.algorithms = {
                k: v(clients)
                for k, v in registered_algorithms.items()
                if (focuses is None or k in focuses)
                and (excludes is None or k not in excludes)
            }
            for v in self.algorithms.values():
                if isinstance(v, IWrapperAlgorithm):
                    v.algorithms = self.algorithms
                v.initialize()
        if self._is_new_http_client:
            self._http_client.start()

    # lifecycle

    async def destroy(self) -> None:
        if self._is_new_http_client:
            await self._http_client.stop()

    # algorithms

    async def _run(
        self, data: BaseModel, endpoint: str, **kw: Any
    ) -> List[Image.Image]:
        if isinstance(data, ReturnArraysModel):
            data.return_arrays = True
        task = endpoint2algorithm(endpoint)
        arrays = await self.algorithms[task].run(data, **kw)
        return list(map(Image.fromarray, arrays))

    async def txt2img(self, data: Txt2ImgSDModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, txt2img_sd_endpoint, **kw)

    async def img2img(self, data: Img2ImgSDModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, img2img_sd_endpoint, **kw)

    async def sr(self, data: Img2ImgSRModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, img2img_sr_endpoint, **kw)

    async def sod(self, data: Img2ImgSODModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, img2img_sod_endpoint, **kw)

    async def inpainting(
        self, data: Img2ImgInpaintingModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, img2img_inpainting_endpoint, **kw)

    async def semantic2img(
        self, data: Img2ImgSemantic2ImgModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, img2img_semantic2img_endpoint, **kw)

    async def sd_inpainting(
        self, data: Txt2ImgSDInpaintingModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, txt2img_sd_inpainting_endpoint, **kw)

    async def sd_outpainting(
        self, data: Txt2ImgSDOutpaintingModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, txt2img_sd_outpainting_endpoint, **kw)

    async def image_captioning(self, data: Img2TxtModel, **kw: Any) -> List[str]:
        task = endpoint2algorithm(img2txt_caption_endpoint)
        result: TextModel = await self.algorithms[task].run(data, **kw)
        return [result.text]

    async def run_multi_controlnet(
        self, data: ControlMultiModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, new_control_multi_endpoint, **kw)

    async def harmonization(
        self, data: Img2ImgHarmonizationModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, img2img_harmonization_endpoint, **kw)

    async def paste_pipeline(
        self, data: PastePipelineModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, paste_pipeline_endpoint, **kw)

    async def blur(self, data: BlurModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_blur_endpoint, **kw)

    async def grayscale(self, data: CVImageModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_grayscale_endpoint, **kw)

    async def erode(self, data: ErodeModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_erode_endpoint, **kw)

    async def resize(self, data: ResizeModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_resize_endpoint, **kw)

    async def affine(self, data: AffineModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_affine_endpoint, **kw)

    async def get_mask(self, data: CVImageModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_get_mask_endpoint, **kw)

    async def inverse(self, data: CVImageModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_inverse_endpoint, **kw)

    async def fill_bg(self, data: FillBGModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_fill_bg_endpoint, **kw)

    async def get_size(self, data: ImageModel, **kw: Any) -> List[int]:
        task = endpoint2algorithm(cv_get_size_endpoint)
        return await self.algorithms[task].run(data, **kw)

    async def modify_box(self, data: ModifyBoxModel, **kw: Any) -> List[List[int]]:
        task = endpoint2algorithm(cv_modify_box_endpoint)
        return await self.algorithms[task].run(data, **kw)

    async def generate_masks(
        self, data: GenerateMasksModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, cv_generate_masks_endpoint, **kw)

    async def crop_image(self, data: CropImageModel, **kw: Any) -> List[Image.Image]:
        return await self._run(data, cv_crop_image_endpoint, **kw)

    async def histogram_match(
        self, data: HistogramMatchModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, cv_histogram_match_endpoint, **kw)

    async def image_similarity(
        self, data: ImageSimilarityModel, **kw: Any
    ) -> List[float]:
        task = endpoint2algorithm(cv_image_similarity_endpoint)
        result: ImageSimilarityResponse = await self.algorithms[task].run(data, **kw)
        return [result.similarity]

    async def repositioning(
        self, data: RepositioningModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, cv_repositioning_endpoint, **kw)

    async def prompt_enhance(self, data: PromptEnhanceModel, **kw: Any) -> List[str]:
        task = endpoint2algorithm(txt2txt_prompt_enhance_endpoint)
        result: PromptEnhanceResponse = await self.algorithms[task].run(data)
        return result.prompts

    async def upscale_tile(
        self, data: UpscaleTileModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, upscale_tile_endpoint, **kw)

    # third party

    async def facexlib_parse(
        self, data: FacexlibParseModel, **kw: Any
    ) -> List[Image.Image]:
        return await self._run(data, facexlib_parse_endpoint, **kw)

    async def facexlib_detect(
        self, data: FacexlibDetectModel, **kw: Any
    ) -> List[List[int]]:
        task = endpoint2algorithm(facexlib_detect_endpoint)
        result: FacexlibDetectResponse = await self.algorithms[task].run(data, **kw)
        return result.ltrbs

    # special

    async def get_image(self, data: ImageModel, **kw: Any) -> List[Image.Image]:
        image = await download_image_with_retry(self._http_client.session, data.url)
        return [image]

    async def add_text(self, data: TextModel, **kw: Any) -> List[str]:
        return [data.text]

    async def get_control_hint(
        self, hint_type: ControlNetHints, data: Dict[str, Any], **kw: Any
    ) -> List[Image.Image]:
        data = control_hint2hint_data_models[hint_type](**data)
        endpoint = control_hint2hint_endpoints[hint_type]
        return await self._run(data, endpoint, **kw)

    async def for_each(self, data: ForEachModel, **kw: Any) -> List[Any]:
        targets = []
        target_key = "$target"
        loop_back_key = "$loop_back"
        params_prefix = "params."
        get_cache_key = lambda k: f"$cache_{k}"
        keys = sorted(data.loops)
        value_lists = [data.loops[k] for k in keys]
        lengths = [len(l) for l in value_lists]
        if len(set(lengths)) != 1:
            msg = f"lengths of value_lists should be equal, got {lengths} (keys={keys})"
            raise ValueError(msg)
        base_caches = {}
        for i, i_values in enumerate(zip(*value_lists)):
            i_kw = shallow_copy_dict(kw)
            for k in list(i_kw):
                if k.startswith(params_prefix):
                    nk = k.lstrip(params_prefix)
                    i_kw[nk] = i_kw.pop(k)
            i_caches = {get_cache_key(k): [v] for k, v in zip(keys, i_values)}
            i_caches.update(shallow_copy_dict(base_caches))
            i_injections = {get_cache_key(k): {"index": 0, "field": k} for k in keys}
            if data.loop_backs is not None and i > 0:
                i_injections[loop_back_key] = data.loop_backs
            i_workflow = Workflow()
            i_workflow.push(
                WorkNode(
                    key=target_key,
                    endpoint=data.endpoint,
                    injections=i_injections,
                    data=shallow_copy_dict(data.params),
                )
            )
            i_results = await self.execute(i_workflow, target_key, i_caches, **i_kw)
            i_target = i_results[target_key]
            targets.append(i_target)
            if data.loop_backs is not None:
                base_caches[loop_back_key] = i_target
        return list(map(list, zip(*targets)))

    async def pick(self, data: PickModel, **kw: Any) -> List[Any]:
        return [data.values[data.index]]

    # workflow

    async def execute(
        self,
        workflow: Workflow,
        target: str,
        caches: Optional[Union[OrderedDict, Dict[str, Any]]] = None,
        *,
        return_if_exception: bool = False,
        **kwargs: Any,
    ) -> Dict[str, TRes]:
        def _inject(
            k: str,
            ki_cache: TRes,
            current_node_data: dict,
            is_for_each: bool = False,
        ) -> None:
            def _set_inject_with_loops_condition() -> None:
                if is_loops:
                    v0[next_k_with_loops_condition] = ki_cache
                else:
                    _inject(next_k_with_loops_condition, ki_cache, v0)

            k_split = k.split(".")
            k0 = k_split[0]
            v0 = current_node_data.get(k0)
            next_k_with_loops_condition = ".".join(k_split[1:])
            is_loops = is_for_each and k0 == "loops"
            if isinstance(v0, dict):
                _set_inject_with_loops_condition()
            elif isinstance(v0, (str, int, float)) or v0 is None:
                if len(k_split) == 1:
                    if isinstance(ki_cache, Image.Image):
                        # we assign random_hash here to make mapping possible
                        current_node_data[k0] = random_hash()
                    else:
                        current_node_data[k0] = ki_cache
                else:
                    if v0 is None:
                        v0 = current_node_data[k0] = {}
                        _set_inject_with_loops_condition()
                    else:
                        raise ValueError(
                            f"field under '{k0}' is already a vanilla value, "
                            f"but further keys are given: '{k_split[1:]}'"
                        )
            elif isinstance(v0, list):
                try:
                    k1 = int(k_split[1])
                except Exception:
                    raise ValueError(f"expected int key for '{k0}', got '{k_split[1]}'")
                v1 = v0[k1]
                if isinstance(v1, dict):
                    _inject(".".join(k_split[2:]), ki_cache, v1)
                elif len(k_split) == 2:
                    if isinstance(ki_cache, Image.Image):
                        # we assign random_hash here to make mapping possible
                        v0[k1] = random_hash()
                    else:
                        v0[k1] = ki_cache
                else:
                    raise ValueError(
                        f"list under '{k0}' is already a vanilla list, "
                        f"but further keys are given: '{k_split[2:]}'"
                    )
            else:
                raise ValueError(
                    f"field under '{k0}' should be one of "
                    "(BaseModel, str, int, float, list), "
                    f"but got '{type(v0)}'"
                )

        all_latencies = {}
        if caches is None:
            caches = OrderedDict()
        else:
            caches = OrderedDict(caches)
            workflow = workflow.copy().inject_caches(caches)
        try:
            for layer in workflow.get_dependency_path(target).hierarchy:
                for item in layer:
                    if item.key in caches:
                        continue
                    node = item.data
                    node_kw = shallow_copy_dict(kwargs)
                    node_data = shallow_copy_dict(node.data)
                    endpoint = node.endpoint
                    is_for_each = endpoint == FOR_EACH_ENDPOINT
                    for k, k_packs in node.injections.items():
                        if not isinstance(k_packs, list):
                            k_packs = [k_packs]
                        for k_pack in k_packs:
                            if k_pack.index is None:
                                ki_cache = caches[k]
                            else:
                                ki_cache = caches[k][k_pack.index]
                            _inject(k_pack.field, ki_cache, node_data, is_for_each)
                            if isinstance(ki_cache, Image.Image):
                                node_kw[k_pack.field] = ki_cache
                    method_fn = getattr(self, endpoint2method[endpoint])
                    t = time.time()
                    if endpoint == UPLOAD_ENDPOINT:
                        data_model = ImageModel(**node_data)
                        item_res = await method_fn(data_model, **node_kw)
                    elif endpoint == ADD_TEXT_ENDPOINT:
                        data_model = TextModel(**node_data)
                        item_res = await method_fn(data_model, **node_kw)
                    elif endpoint == CONTROL_HINT_ENDPOINT:
                        hint_type = node_data.pop("hint_type")
                        item_res = await method_fn(hint_type, node_data, **node_kw)
                        endpoint = control_hint2hint_endpoints[hint_type]
                    elif endpoint == FOR_EACH_ENDPOINT:
                        data_model = ForEachModel(**node_data)
                        item_res = await method_fn(data_model, **node_kw)
                    elif endpoint == PICK_ENDPOINT:
                        data_model = PickModel(**node_data)
                        item_res = await method_fn(data_model, **node_kw)
                    else:
                        data_model = self.get_data_model(endpoint, node_data)
                        item_res = await method_fn(data_model, **node_kw)
                    if endpoint == UPLOAD_ENDPOINT or endpoint == ADD_TEXT_ENDPOINT:
                        ls = dict(download=time.time() - t)
                    elif endpoint == FOR_EACH_ENDPOINT:
                        ls = dict(loop=time.time() - t)
                    elif endpoint == PICK_ENDPOINT:
                        ls = dict(pick=time.time() - t)
                    else:
                        ls = self.algorithms[
                            endpoint2algorithm(endpoint)
                        ].last_latencies
                    caches[item.key] = item_res
                    all_latencies[item.key] = ls
            caches[EXCEPTION_MESSAGE_KEY] = None
        except Exception as err:
            caches[EXCEPTION_MESSAGE_KEY] = get_err_msg(err)
            if not return_if_exception:
                raise err
        caches[ALL_LATENCIES_KEY] = all_latencies
        return caches

    # misc

    def get_data_model(self, endpoint: str, data: Dict[str, Any]) -> BaseModel:
        task = endpoint2algorithm(endpoint)
        return self.algorithms[task].model_class(**data)


__all__ = [
    "UPLOAD_ENDPOINT",
    "ADD_TEXT_ENDPOINT",
    "CONTROL_HINT_ENDPOINT",
    "FOR_EACH_ENDPOINT",
    "PICK_ENDPOINT",
    "ALL_LATENCIES_KEY",
    "APIs",
    "HighresModel",
    "Img2TxtModel",
    "Txt2ImgSDModel",
    "Img2ImgSDModel",
    "Img2ImgSRModel",
    "Img2ImgSODModel",
    "Img2ImgInpaintingModel",
    "Txt2ImgSDInpaintingModel",
    "Txt2ImgSDOutpaintingModel",
    "ControlNetHints",
    "ControlMultiModel",
    "Img2ImgHarmonizationModel",
]



import os
import json
import time
import yaml
import redis
import torch
import asyncio
import datetime
import logging.config

import numpy as np

from obs import *
from PIL import Image
from kafka import KafkaConsumer
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union
from typing import Callable
from fastapi import Response
from pydantic import BaseModel

from cfclient.models import *
from cftool.misc import get_err_msg
from cftool.misc import shallow_copy_dict
from cfclient.core import HttpClient
from cfclient.core import TritonClient
from cfclient.utils import run_algorithm
from huaweicloudsdkcore.auth.credentials import BasicCredentials
from huaweicloudsdkmoderation.v2 import ModerationClient
from huaweicloudsdkmoderation.v2.region.moderation_region import ModerationRegion

# This is necessary to register the algorithms
from cfcreator import *
from cfcreator.legacy.control_multi import ControlMultiModel as LegacyControlMultiModel

from producer import dump_queue
from producer import check_timeout
from producer import StatusData


# logging
root = os.path.dirname(__file__)
logging_root = os.path.join(root, "logs", "consumer")
os.makedirs(logging_root, exist_ok=True)
with open(os.path.join(root, "config.yml")) as f:
    now = datetime.datetime.now()
    timestamp = now.strftime("%Y-%m-%d_%H-%M-%S-%f")
    log_path = os.path.join(logging_root, f"{timestamp}.log")
    config = yaml.load(f, Loader=yaml.FullLoader)
    config["handlers"]["file"]["filename"] = log_path
    logging.config.dictConfig(config)

logging.getLogger("aiokafka").disabled = True
logging.getLogger("kafka.conn").disabled = True
logging.getLogger("kafka.client").disabled = True
logging.getLogger("kafka.cluster").disabled = True
logging.getLogger("kafka.coordinator").disabled = True
logging.getLogger("kafka.coordinator.consumer").disabled = True
logging.getLogger("kafka.metrics.metrics").disabled = True
logging.getLogger("kafka.protocol.parser").disabled = True
logging.getLogger("kafka.consumer.fetcher").disabled = True
logging.getLogger("kafka.consumer.group_coordinator").disabled = True
logging.getLogger("kafka.consumer.subscription_state").disabled = True


constants = dict(
    triton_host=None,
    triton_port=8000,
)


# clients
## cos client
cos_client = ObsClient(access_key_id=AK, secret_access_key=SK, server=SERVER)
image_mod_cred = BasicCredentials(AK, SK)
image_mod_client = (
    ModerationClient.new_builder()
    .with_credentials(image_mod_cred)
    .with_region(ModerationRegion.value_of(REGION))
    .build()
)
## http client
http_client = HttpClient()
## triton client
triton_host = constants["triton_host"]
if triton_host is None:
    triton_client = None
else:
    triton_client = TritonClient(url=f"{triton_host}:{constants['triton_port']}")
## collect
clients = dict(
    http=http_client,
    triton=triton_client,
)

redis_client = redis.Redis(**redis_kwargs())
pending_queue_key = get_pending_queue_key()


# algorithms
loaded_algorithms: Dict[str, IAlgorithm] = {
    k: v(clients) for k, v in algorithms.items()
}


def get_redis_number(key: str) -> int:
    data = redis_client.get(key)
    if data is None:
        return 0
    return int(data.decode())  # type: ignore


def get_pending_queue() -> list:
    data = redis_client.get(pending_queue_key)
    if data is None:
        return []
    return json.loads(data)


async def post_callback(
    url: str,
    uid: str,
    success: bool,
    data: Dict[str, Any],
    retry: int = 10,
    interval: int = 3,
) -> None:
    if not url:
        print(">>>>> no callback is required")
        return

    print(f">>>>> post callback to {url}")

    async def fn() -> None:
        cb_data = dict(uid=uid, success=success, data=data)
        async with http_client.session.post(url, json=cb_data, timeout=interval) as res:
            if not res.ok:
                raise ValueError(f"post callback failed ({res.status})")
            print(f">>>>> post callback succeeded ({await res.json()})")

    msg = ""
    for i in range(retry):
        try:
            await fn()
            if i > 0:
                logging.warning(f"succeeded after {i} retries ({msg})")
            return
        except Exception as err:
            msg = get_err_msg(err)
        time.sleep(interval)
    if msg:
        print(
            f"\n\n!!! post to callback_url ({url}) failed "
            f"(After {retry} retries) ({msg}) !!!\n\n"
        )


def simplify(params: Any) -> Any:
    def _core(p: Any) -> Any:
        if isinstance(p, list):
            return [_core(v) for v in p]
        if not isinstance(p, dict):
            return p
        p = shallow_copy_dict(p)
        for k, v in p.items():
            if k == "custom_embeddings":
                if v is None or not isinstance(v, dict):
                    continue
                simplified_embeddings = {}
                for vk, vv in v.items():
                    if vv is None:
                        continue
                    try:
                        vva = np.atleast_2d(vv)
                        if vva.shape[1] <= 6:
                            simplified_embeddings[vk] = vv
                        else:
                            vva = np.concatenate([vva[:, :3], vva[:, -3:]], axis=1)
                            simplified_embeddings[vk] = vva.tolist()
                    except Exception as err:
                        simplified_embeddings[vk] = get_err_msg(err)
                p[k] = simplified_embeddings
            elif isinstance(p, (list, dict)):
                p[k] = _core(v)
        return p

    return _core(params)


def get_upload_results(raw: List[Any], upload_fn: Callable) -> List[Any]:
    def _upload(v: List[Any]) -> List[Any]:
        return [
            _upload(elem)
            if isinstance(elem, list)
            else upload_fn(cos_client, elem)
            if isinstance(elem, np.ndarray)
            else upload_fn(cos_client, np.array(elem))
            if isinstance(elem, Image.Image)
            else elem
            for elem in v
        ]

    return _upload(raw)


def extract_urls(results: List[Union[UploadResponse, Any]], attr: str) -> List[Any]:
    def _extract(v: List[Any]) -> List[Any]:
        return [
            _extract(elem)
            if isinstance(elem, list)
            else getattr(elem, attr)
            if isinstance(elem, UploadResponse)
            else elem
            for elem in v
        ]

    return _extract(results)


# return (urls, reasons)
def audit_urls(
    model: BaseModel,
    url_results: List[Union[UploadResponse, Any]],
) -> Tuple[List[str], List[str]]:
    urls = extract_urls(url_results, "cdn")
    if isinstance(model, UseAuditModel) and not model.use_audit:
        reasons = [""] * len(url_results)
    else:
        reasons = []
        for i, rs in enumerate(url_results):
            if not isinstance(rs, UploadResponse):
                reasons.append("")
                continue
            try:
                audit = audit_image(cos_client, image_mod_client, rs.path)
            except:
                audit = AuditResponse(safe=False, reason="unknown")
            if audit.safe:
                reasons.append("")
            else:
                urls[i] = ""
                reasons.append(audit.reason)
    return urls, reasons


# kafka & redis
async def consume() -> None:
    OPT["verbose"] = False
    OPT["lazy_load"] = True

    topic = kafka_topic()
    expire_seconds = 10 * 365 * 24 * 3600

    redis_client.expire(pending_queue_key, expire_seconds)
    # initialize
    http_client.start()
    for v in loaded_algorithms.values():
        if isinstance(v, IWrapperAlgorithm):
            v.algorithms = loaded_algorithms
        v.initialize()
    print("> Algorithms are Loaded!")
    kafka_consumer = KafkaConsumer(
        topic,
        group_id=kafka_group_id(),
        bootstrap_servers=kafka_server(),
        max_poll_records=kafka_max_poll_records(),
        max_poll_interval_ms=kafka_max_poll_interval_ms(),
    )
    with open("/tmp/health", "w"):
        pass
    # main loop
    try:
        for message in kafka_consumer:
            data = json.loads(message.value)
            uid = data["uid"]
            task = data["task"]
            if task == "scene-generation":
                task = "txt2img.sd.outpainting"
            params = data["params"]
            callback_url = params.get("callback_url", "")
            existing = redis_client.get(uid)
            if existing is not None:
                existing = json.loads(existing)
                print(">>> existing", existing)
                if existing["status"] in (
                    Status.FINISHED,
                    Status.EXCEPTION,
                    Status.INTERRUPTED,
                ):
                    continue
            if existing is not None and check_timeout(uid, StatusData(**existing)):
                print("!!! timeout", uid)
                continue
            print(">>> working", uid)
            data = {} if existing is None else (existing.get("data", {}) or {})
            start_time = time.time()
            data["start_time"] = start_time
            create_time = data.get("create_time", start_time)
            redis_client.set(uid, json.dumps(dict(status=Status.WORKING, data=data)))
            # maintain queue
            queue = get_pending_queue()
            if uid not in queue:
                queue.insert(0, uid)
                dump_queue(queue)
            procedure = "start"
            model = None
            try:
                algorithm = loaded_algorithms[task]
                model = algorithm.model_class(**params)  # type: ignore
                procedure = "start -> run_algorithm"
                if isinstance(model, ReturnArraysModel):
                    model.return_arrays = True
                res: Union[Response, Any] = await run_algorithm(algorithm, model)
                latencies = algorithm.last_latencies
                t1 = time.time()
                if isinstance(algorithm, WorkflowAlgorithm):
                    intermediate = {}
                    response = dict(
                        is_exception=res.pop(WORKFLOW_IS_EXCEPTION_KEY, False),
                        intermediate=intermediate,
                    )
                    result = dict(uid=uid, response=response)
                    all_results = {}
                    save_to_private = (
                        not isinstance(model, WorkflowModel)
                        or model.save_intermediate_to_private
                    )
                    for k, v in res.items():
                        procedure = f"[{k}] run_algorithm -> upload_temp_image"
                        upload_fn = (
                            upload_temp_image
                            if k == WORKFLOW_TARGET_RESPONSE_KEY or not save_to_private
                            else upload_private_image
                        )
                        all_results[k] = get_upload_results(v, upload_fn)
                    t2 = time.time()
                    for k, k_results in all_results.items():
                        procedure = f"[{k}] upload_temp_image -> audit_image"
                        if save_to_private and k != WORKFLOW_TARGET_RESPONSE_KEY:
                            intermediate[k] = dict(
                                urls=extract_urls(k_results, "cos"),
                                reasons=[""] * len(k_results),
                            )
                        else:
                            k_urls, k_reasons = audit_urls(model, k_results)
                            if k != WORKFLOW_TARGET_RESPONSE_KEY:
                                intermediate[k] = dict(urls=k_urls, reasons=k_reasons)
                            else:
                                response["urls"] = k_urls
                                response["reasons"] = k_reasons
                    t3 = time.time()
                    procedure = "audit_image -> redis"
                elif task.startswith("control") or task.startswith("pipeline"):
                    procedure = "run_algorithm -> upload_temp_image"
                    url_results = get_upload_results(res, upload_temp_image)
                    t2 = time.time()
                    procedure = "upload_temp_image -> audit_image"
                    urls, reasons = audit_urls(model, url_results)
                    t3 = time.time()
                    procedure = "audit_image -> redis"
                    if isinstance(algorithm, WorkflowAlgorithm):
                        result = dict(
                            uid=uid,
                            response=dict(urls=urls, reasons=reasons),
                        )
                    elif task.startswith("control"):
                        if isinstance(model, ControlMultiModel):
                            keys = set(get_bundle_key(b) for b in model.controls)
                            num_cond = len(keys)
                        elif isinstance(model, LegacyControlMultiModel):
                            num_cond = len(model.types)
                        else:
                            num_cond = 1
                        result = dict(
                            uid=uid,
                            response=dict(
                                hint_urls=urls[:num_cond],
                                hint_reasons=reasons[:num_cond],
                                result_urls=urls[num_cond:],
                                result_reasons=reasons[num_cond:],
                            ),
                        )
                    elif task == "pipeline.paste" or task == "pipeline.product_sr":
                        result = dict(
                            uid=uid,
                            response=dict(url=urls[0], reason=reasons[0]),
                        )
                    elif task == "pipeline.product":
                        result = dict(
                            uid=uid,
                            response=dict(
                                raw_url=urls[0],
                                raw_reason=reasons[0],
                                depth_url=urls[1],
                                depth_reason=reasons[1],
                                result_urls=urls[2:],
                                result_reasons=reasons[2:],
                            ),
                        )
                    elif task.startswith("pipeline"):
                        num_hints = len(params["types"])
                        result = dict(
                            uid=uid,
                            response=dict(
                                raw_url=urls[0],
                                raw_reason=reasons[0],
                                hint_urls=urls[1 : num_hints + 1],
                                hint_reasons=reasons[1 : num_hints + 1],
                                result_urls=urls[num_hints + 1 :],
                                result_reasons=reasons[num_hints + 1 :],
                            ),
                        )
                    else:
                        raise ValueError(f"unrecognized task '{task}' occurred")
                elif algorithm.response_model_class is not None:
                    t2 = t3 = t1
                    procedure = "run_algorithm -> redis"
                    result = dict(uid=uid, response=res.dict())
                else:
                    procedure = "run_algorithm -> upload_temp_image"
                    if isinstance(res, list) and not res:
                        t2 = t3 = time.time()
                        result = dict(uid=uid, cdn="", cos="", safe=True, reason="")
                    else:
                        content = res[0] if isinstance(res, list) else res.body
                        urls = upload_temp_image(cos_client, content)
                        t2 = time.time()
                        procedure = "upload_temp_image -> audit_image"
                        if task != "img2img.sr":
                            try:
                                audit = audit_image(
                                    cos_client, image_mod_client, urls.path
                                )
                            except:
                                audit = AuditResponse(safe=False, reason="unknown")
                        else:
                            audit = AuditResponse(safe=True, reason="")
                        t3 = time.time()
                        procedure = "audit_image -> redis"
                        result = dict(
                            uid=uid,
                            cdn=urls.cdn if audit.safe else "",
                            cos=urls.cos if audit.safe else "",
                            safe=audit.safe,
                            reason=audit.reason,
                        )
                result.update(data)
                end_time = time.time()
                result["end_time"] = end_time
                result["duration"] = end_time - create_time
                result["elapsed_times"] = dict(
                    pending=start_time - create_time,
                    run_algorithm=t1 - start_time,
                    algorithm_latencies=latencies,
                    upload=t2 - t1,
                    audit=t3 - t2,
                )
                result["request"] = dict(task=task, model=simplify(model.dict()))
                redis_client.set(
                    uid,
                    json.dumps(dict(status=Status.FINISHED, data=result)),
                )
                t4 = time.time()
                procedure = "redis -> callback"
                await post_callback(callback_url, uid, True, result)
                t5 = time.time()
                procedure = "callback -> update_elapsed_times"
                result["elapsed_times"].update(
                    dict(
                        redis=t4 - t3,
                        callback=t5 - t4,
                    )
                )
                redis_client.set(
                    uid,
                    json.dumps(dict(status=Status.FINISHED, data=result)),
                )
                procedure = "done"
                # maintain queue
                queue = get_pending_queue()
                if uid in queue:
                    queue.remove(uid)
                    dump_queue(queue)
            except Exception as err:
                end_time = time.time()
                torch.cuda.empty_cache()
                reason = f"{task} -> {procedure} : {get_err_msg(err)}"
                logging.exception(reason)
                data["uid"] = uid
                data["reason"] = reason
                data["end_time"] = end_time
                data["duration"] = end_time - create_time
                if model is None:
                    data["request"] = dict(task=task, params=simplify(params))
                else:
                    data["request"] = dict(task=task, model=simplify(model.dict()))
                redis_client.set(
                    uid,
                    json.dumps(
                        dict(status=Status.EXCEPTION, data=data),
                        ensure_ascii=False,
                    ),
                )
                await post_callback(callback_url, uid, False, data)
    finally:
        # clean up
        await http_client.stop()
        print(">>> end")


asyncio.run(consume())



from cfcreator.endpoints import *
from cfcreator.sdks.apis import *


workflow = Workflow()
workflow.push(
    WorkNode(
        key="txt2img_0",
        endpoint=txt2img_sd_endpoint,
        injections={},
        data=dict(text="A lovely little girl.", seed="123", version="anime_anything"),
    )
)
workflow.push(
    WorkNode(
        key="img2img_0",
        endpoint=img2img_sd_endpoint,
        injections=dict(
            txt2img_0=dict(index=0, field="url"),
        ),
        data=dict(
            text="A lovely little girl.",
            seed="345",
            version="anime_anything",
            fidelity=0.6,
        ),
    )
)
workflow.push(
    WorkNode(
        key="controlnet_0",
        endpoint=new_control_multi_endpoint,
        injections=dict(
            txt2img_0=dict(index=0, field="controls.0.data.hint_url"),
            img2img_0=dict(index=0, field="controls.1.data.hint_url"),
        ),
        data=dict(
            prompt="A lovely little girl.",
            seed="567",
            base_model="anime_anything",
            controls=[
                dict(type="depth", data=dict(control_strength=0.8)),
                dict(type="canny", data=dict(control_strength=0.6)),
            ],
        ),
    )
)


if __name__ == "__main__":
    import os
    import asyncio

    apis = APIs(
        focuses_endpoints=[
            txt2img_sd_endpoint,
            img2img_sd_endpoint,
            new_control_multi_endpoint,
        ]
    )
    target = "controlnet_0"
    results = asyncio.run(apis.execute(workflow, target))
    output_folder = "workflow"
    os.makedirs(output_folder, exist_ok=True)
    for i, (k, res) in enumerate(results.items()):
        for j, image in enumerate(res):
            image.save(os.path.join(output_folder, f"{i}_{k}_{j}.png"))
    asyncio.run(apis.destroy())



import time

import numpy as np

from PIL import Image
from PIL import ImageDraw
from typing import Any
from typing import List
from typing import Tuple
from typing import Optional
from fastapi import Response
from pydantic import Field
from cftool.misc import random_hash
from cftool.misc import shallow_copy_dict
from cftool.types import TNumberPair

from .common import get_response
from .common import Txt2ImgModel
from .common import InpaintingMode
from .common import IWrapperAlgorithm
from .common import ReturnArraysModel
from .control_multi import TBundle
from .control_multi import ControlMultiModel


upscale_tile_endpoint = "/upscale/tile"


class UpscaleTileModel(ReturnArraysModel, Txt2ImgModel):
    url: str = Field(..., description="url of the initial image")
    padding: int = Field(32, description="padding for each tile")
    grid_wh: TNumberPair = Field(None, description="explicit specify the grid size")
    upscale_factor: TNumberPair = Field(2, ge=1, description="upscale factor")
    fidelity: float = Field(0.45, description="fidelity of each tile")
    highres_steps: int = Field(36, description="num_steps for upscaling")
    strength: float = Field(1.0, description="strength of the tile control")
    controls: Optional[List[TBundle]] = Field(None, description="Extra controls")


def resize(image: Image.Image, size: Tuple[int, int]) -> Image.Image:
    return image.resize(size, Image.LANCZOS)


@IWrapperAlgorithm.auto_register()
class UpscaleTile(IWrapperAlgorithm):
    model_class = UpscaleTileModel

    endpoint = upscale_tile_endpoint

    async def run(self, data: UpscaleTileModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        canvas = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        if data.grid_wh is None:
            w_grid, h_grid = canvas.size
        else:
            grid_wh = data.grid_wh
            if isinstance(grid_wh, int):
                grid_wh = grid_wh, grid_wh
            w_grid, h_grid = grid_wh
        factor = data.upscale_factor
        if isinstance(factor, int):
            factor = factor, factor
        w_factor, h_factor = factor
        w, h = w_grid * w_factor, h_grid * h_factor
        canvas = resize(canvas, (w, h))
        for k, v in kwargs.items():
            if k != "url" and isinstance(v, Image.Image):
                kwargs[k] = resize(v, (w, h))
        all_black = Image.new("RGB", (w, h), color=(0, 0, 0))
        all_black_draw = ImageDraw.Draw(all_black)
        controlnet_data = ControlMultiModel(
            url=random_hash(),
            mask_url=random_hash(),
            prompt=data.text,
            base_model=data.version,
            seed=data.seed,
            sampler=data.sampler,
            inpainting_mode=InpaintingMode.MASKED,
            inpainting_target_wh=(w_grid, h_grid),
            controls=[
                dict(
                    type="control_v11f1e_sd15_tile",
                    data=dict(
                        hint_url=random_hash(),
                        control_strength=data.strength,
                    ),
                )
            ],
            keep_original=False,
            # params
            num_steps=data.highres_steps,
            use_reference=True,
            reference_fidelity=data.fidelity,
            inpainting_mask_padding=data.padding,
        )
        if data.controls is not None:
            controlnet_data.controls = data.controls + controlnet_data.controls
        n_controls = len(controlnet_data.controls)
        t2 = time.time()
        for j in range(h_factor):
            jy = j % h_factor * h_grid
            for i in range(w_factor):
                ix = i % w_factor * w_grid
                lt_rb = ix, jy, ix + w_grid, jy + h_grid
                all_black_draw.rectangle(lt_rb, fill=(255, 255, 255))
                kw = shallow_copy_dict(kwargs)
                kw.update(
                    {
                        "url": canvas,
                        "mask_url": all_black,
                        f"controls.{n_controls - 1}.data.hint_url": canvas,
                    }
                )
                images = await self.apis.run_multi_controlnet(controlnet_data, **kw)
                canvas = images[-1]
                all_black_draw.rectangle(lt_rb, fill=(0, 0, 0))
        t3 = time.time()
        res = get_response(data, [np.array(canvas)])
        latencies = {
            "download": t1 - t0,
            "preprocess": t2 - t1,
            "inference": t3 - t2,
            "get_response": time.time() - t3,
        }
        self.log_times(latencies)
        return res


__all__ = [
    "upscale_tile_endpoint",
    "UpscaleTileModel",
    "UpscaleTile",
]



from cfcreator import *
from cfcreator.sdks import *


async def main():
    sdk = DirectSDK()
    model = Txt2ImgSDModel(text="A lovely little cat.", num_steps=5)
    img = await sdk.txt2img(model)
    img.save("txt2img.png")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



import time

import numpy as np

from PIL import Image
from enum import Enum
from typing import Any
from typing import Optional
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cftool.cv import to_uint8
from cfclient.models import ImageModel

from .utils import APIs
from .common import register_sd
from .common import register_sd_inpainting
from .common import get_sd_from
from .common import get_response
from .common import handle_diffusion_model
from .common import get_normalized_arr_from_diffusion
from .common import handle_diffusion_inpainting_model
from .common import IAlgorithm
from .common import HighresModel
from .common import Txt2ImgModel
from .common import ReturnArraysModel
from .common import CommonSDInpaintingModel


txt2img_sd_endpoint = "/txt2img/sd"
txt2img_sd_inpainting_endpoint = "/txt2img/sd.inpainting"
txt2img_sd_outpainting_endpoint = "/txt2img/sd.outpainting"


class _Txt2ImgSDModel(BaseModel):
    w: int = Field(512, description="The desired output width.")
    h: int = Field(512, description="The desired output height.")
    highres_info: Optional[HighresModel] = Field(None, description="Highres info.")


class Txt2ImgSDModel(ReturnArraysModel, Txt2ImgModel, _Txt2ImgSDModel):
    pass


@IAlgorithm.auto_register()
class Txt2ImgSD(IAlgorithm):
    model_class = Txt2ImgSDModel

    endpoint = txt2img_sd_endpoint

    def initialize(self) -> None:
        register_sd()

    async def run(self, data: Txt2ImgSDModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        m = get_sd_from(APIs.SD, data)
        t1 = time.time()
        size = data.w, data.h
        kwargs.update(handle_diffusion_model(m, data))
        if data.highres_info is not None:
            kwargs["highres_info"] = data.highres_info.dict()
        img_arr = m.txt2img(
            data.text,
            size=size,
            max_wh=data.max_wh,
            **kwargs,
        ).numpy()[0]
        t2 = time.time()
        res = get_response(data, [to_uint8(get_normalized_arr_from_diffusion(img_arr))])
        self.log_times(
            {
                "get_model": t1 - t0,
                "inference": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


class PaddingModes(str, Enum):
    CV2_NS = "cv2_ns"
    CV2_TELEA = "cv2_telea"


class Txt2ImgSDInpaintingModel(CommonSDInpaintingModel, Txt2ImgModel, ImageModel):
    mask_url: str = Field(
        ...,
        description="""
The `cdn` / `cos` url of the user's mask.
> `cos` url from `qcloud` is preferred.
> If empty string is provided, then we will use an empty mask, which means we will simply perform an image-to-image transform.  
""",
    )


class Txt2ImgSDOutpaintingModel(CommonSDInpaintingModel, Txt2ImgModel, ImageModel):
    pass


@IAlgorithm.auto_register()
class Txt2ImgSDInpainting(IAlgorithm):
    model_class = Txt2ImgSDInpaintingModel

    endpoint = txt2img_sd_inpainting_endpoint

    def initialize(self) -> None:
        register_sd()
        register_sd_inpainting()

    async def run(
        self,
        data: Txt2ImgSDInpaintingModel,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        mask = await self.get_image_from("mask_url", data, kwargs)
        t1 = time.time()
        if data.use_raw_inpainting:
            api_key = APIs.SD
        else:
            api_key = APIs.SD_INPAINTING
        m = get_sd_from(api_key, data)
        t2 = time.time()
        kwargs.update(handle_diffusion_model(m, data))
        kwargs.update(handle_diffusion_inpainting_model(data))
        mask_arr = np.array(mask)
        mask_arr[..., -1] = np.where(mask_arr[..., -1] > 0, 255, 0)
        mask = Image.fromarray(mask_arr)
        img_arr = m.txt2img_inpainting(data.text, image, mask, **kwargs).numpy()[0]
        t3 = time.time()
        res = get_response(data, [to_uint8(get_normalized_arr_from_diffusion(img_arr))])
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "get_response": time.time() - t3,
            }
        )
        return res


@IAlgorithm.auto_register()
class Txt2ImgSDOutpainting(IAlgorithm):
    model_class = Txt2ImgSDOutpaintingModel

    endpoint = txt2img_sd_outpainting_endpoint

    def initialize(self) -> None:
        register_sd_inpainting()

    async def run(
        self,
        data: Txt2ImgSDOutpaintingModel,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        m = get_sd_from(APIs.SD_INPAINTING, data)
        m.disable_control()
        t2 = time.time()
        kwargs.update(handle_diffusion_model(m, data))
        kwargs.update(handle_diffusion_inpainting_model(data))
        img_arr = m.outpainting(data.text, image, **kwargs).numpy()[0]
        t3 = time.time()
        res = get_response(data, [to_uint8(get_normalized_arr_from_diffusion(img_arr))])
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "get_response": time.time() - t3,
            }
        )
        return res


__all__ = [
    "txt2img_sd_endpoint",
    "txt2img_sd_inpainting_endpoint",
    "txt2img_sd_outpainting_endpoint",
    "Txt2ImgSDModel",
    "Txt2ImgSDInpaintingModel",
    "Txt2ImgSDOutpaintingModel",
    "Txt2ImgSD",
    "Txt2ImgSDInpainting",
    "Txt2ImgSDOutpainting",
]



import time

from typing import Any
from typing import List
from typing import Union
from fastapi import Response
from pydantic import Field
from pydantic import Extra
from pydantic import BaseModel
from cftool.cv import np_to_bytes

from .utils import to_canvas
from .common import register_sd
from .common import register_sd_inpainting
from .common import IAlgorithm
from .common import ControlNetModel
from .control import apply_control
from .control import _MLSDModel
from .control import _PoseModel
from .control import _CannyModel
from .control import _DepthModel
from .control import ControlNetHints
from .control import ControlNetBundleCommonData
from .control import ControlNetBundleCommonDataWithDetectResolution


new_control_multi_endpoint = "/control_new/multi"


class MLSDBundleData(_MLSDModel, ControlNetBundleCommonData):
    pass


class MLSDBundle(BaseModel):
    type: ControlNetHints = Field(ControlNetHints.MLSD, const=True)
    data: MLSDBundleData


class PoseBundleData(_PoseModel, ControlNetBundleCommonData):
    pass


class PoseBundle(BaseModel):
    type: ControlNetHints = Field(ControlNetHints.POSE, const=True)
    data: PoseBundleData


class CannyBundleData(_CannyModel, ControlNetBundleCommonData):
    pass


class CannyBundle(BaseModel):
    type: ControlNetHints = Field(ControlNetHints.CANNY, const=True)
    data: CannyBundleData


class DepthBundleData(_DepthModel, ControlNetBundleCommonData):
    pass


class DepthBundle(BaseModel):
    type: ControlNetHints = Field(ControlNetHints.DEPTH, const=True)
    data: DepthBundleData


class UniversalControlModel(
    ControlNetBundleCommonDataWithDetectResolution,
    extra=Extra.allow,
):
    pass


class UniversalBundle(BaseModel):
    type: str
    data: UniversalControlModel


TBundle = Union[MLSDBundle, PoseBundle, CannyBundle, DepthBundle, UniversalBundle]


class ControlMultiModel(ControlNetModel):
    controls: List[TBundle] = Field(..., description="Target controls")


@IAlgorithm.auto_register()
class ControlMulti(IAlgorithm):
    model_class = ControlMultiModel

    endpoint = new_control_multi_endpoint

    def initialize(self) -> None:
        register_sd()
        register_sd_inpainting()

    async def run(self, data: ControlMultiModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        results, latencies = await apply_control(self, data, data.controls, **kwargs)
        t0 = time.time()
        content = None if data.return_arrays else np_to_bytes(to_canvas(results))
        t1 = time.time()
        latencies["to_canvas"] = t1 - t0
        self.log_times(latencies)
        if data.return_arrays:
            return results
        return Response(content=content, media_type="image/png")


__all__ = [
    "new_control_multi_endpoint",
    "ControlMultiModel",
    "ControlMulti",
]



import os
import cv2
import time
import torch
import asyncio

import numpy as np

from PIL import Image
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Optional
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cftool.cv import to_rgb
from cftool.cv import to_uint8
from cftool.cv import np_to_bytes
from cftool.cv import restrict_wh
from cftool.cv import get_suitable_size
from cftool.misc import shallow_copy_dict
from cfclient.models.core import ImageModel
from cflearn.api.cv.diffusion import ControlNetHints
from cflearn.models.cv.diffusion.utils import CONTROL_HINT_KEY
from cflearn.models.cv.diffusion.utils import CONTROL_HINT_END_KEY
from cflearn.models.cv.diffusion.utils import CONTROL_HINT_START_KEY

from .utils import api_pool
from .utils import to_canvas
from .utils import resize_image
from .utils import to_contrast_rgb
from .utils import APIs
from .common import BaseSDTag
from .common import get_sd_from
from .common import register_sd
from .common import register_sd_inpainting
from .common import handle_diffusion_model
from .common import handle_diffusion_inpainting_model
from .common import IAlgorithm
from .common import ControlNetModel
from .common import ReturnArraysModel
from .common import _ControlNetCoreModel


root = os.path.dirname(__file__)
new_control_depth_endpoint = "/control_new/depth"
new_control_canny_endpoint = "/control_new/canny"
new_control_pose_endpoint = "/control_new/pose"
new_control_mlsd_endpoint = "/control_new/mlsd"
control_depth_hint_endpoint = "/control/depth_hint"
control_canny_hint_endpoint = "/control/canny_hint"
control_pose_hint_endpoint = "/control/pose_hint"
control_mlsd_hint_endpoint = "/control/mlsd_hint"


images_type = Tuple[np.ndarray, np.ndarray]
apply_response = Tuple[List[np.ndarray], Dict[str, float]]


class ControlNetBundleCommonData(_ControlNetCoreModel):
    pass


class ControlNetBundleCommonDataWithDetectResolution(ControlNetBundleCommonData):
    detect_resolution: Optional[int] = Field(None, description="Detect resolution.")


class ControlNetBundle(BaseModel):
    type: Union[str, ControlNetHints]
    data: ControlNetBundleCommonDataWithDetectResolution


TKey = Tuple[str, str, bool]


def get_bundle_key(bundle: ControlNetBundle) -> TKey:
    return bundle.type, bundle.data.hint_url, bundle.data.bypass_annotator


def get_hint_url_key(url: str) -> str:
    return f"{url}-hint"


async def apply_control(
    self: IAlgorithm,
    common: ControlNetModel,
    controls: List[ControlNetBundle],
    **kwargs: Any,
) -> apply_response:
    t_sd = time.time()
    api_key = APIs.SD_INPAINTING if common.use_inpainting else APIs.SD
    api = get_sd_from(api_key, common, no_change=True)
    t_sd2 = time.time()
    need_change_device = api_pool.need_change_device(api_key)
    api.enable_control()
    # download images
    urls = set()
    is_hint = {}
    url_to_image = {}
    if common.url is not None:
        urls.add(common.url)
        url_to_image[common.url] = kwargs.get("url")
    if common.mask_url is not None:
        urls.add(common.mask_url)
        url_to_image[common.mask_url] = kwargs.get("mask_url")
    for i, bundle in enumerate(controls):
        if not bundle.data.hint_url:
            raise ValueError("hint url should be provided in `controls`")
        urls.add(bundle.data.hint_url)
        url_to_image[bundle.data.hint_url] = kwargs.get(f"controls.{i}.data.hint_url")
        is_hint[bundle.data.hint_url] = True
    sorted_urls = sorted(urls)
    remained = [i for i, url in enumerate(sorted_urls) if url_to_image[url] is None]
    futures = [self.download_image_with_retry(sorted_urls[i]) for i in remained]
    remained_images: List[Image.Image] = await asyncio.gather(*futures)
    images: List[Image.Image] = [url_to_image[url] for url in sorted_urls]
    for i, index in enumerate(remained):
        images[index] = remained_images[i]
    ## make sure that every image should have the same size
    original_w, original_h = images[0].size
    for im in images[1:]:
        w, h = im.size
        if w != original_w or h != original_h:
            msg = f"image size mismatch: {(original_w, original_h)} vs {(w, h)}"
            raise ValueError(msg)
    ## construct a lookup table
    image_array_d: Dict[str, np.ndarray] = {}
    for url, image in zip(sorted_urls, images):
        if url == common.mask_url:
            image_array_d[url] = np.array(image)
        if url == common.url:
            if url == common.mask_url:
                raise ValueError("`url` and `mask_url` should be different")
            image_array_d[url] = np.array(to_rgb(image))
        if is_hint.get(url, False):
            image_array_d[get_hint_url_key(url)] = np.array(to_contrast_rgb(image))
    # gather detect resolution & extra annotator params
    detect_resolutions = []
    for bundle in controls:
        detect_resolutions.append(getattr(bundle.data, "detect_resolution", None))
    extra_annotator_params_list = []
    for bundle in controls:
        extra_annotator_params_list.append(
            getattr(bundle.data, "extra_annotator_params", None)
        )
    # calculate suitable size
    resize_to_original = lambda array, binarize: cv2.resize(
        array,
        (original_w, original_h),
        interpolation=cv2.INTER_NEAREST if binarize else cv2.INTER_CUBIC,
    )
    w, h = restrict_wh(original_w, original_h, common.max_wh)
    w = get_suitable_size(w, 64)
    h = get_suitable_size(h, 64)
    # activate corresponding ControlNet
    t0 = time.time()
    hint_types = sorted(set(bundle.type for bundle in controls))
    base_md = api.sd_weights.get(BaseSDTag) if common.no_switch else None
    api.switch_control(*hint_types, base_md=base_md)
    t1 = time.time()
    # gather hints
    all_keys: List[TKey] = []
    ## Tensor for input, ndarray for results
    all_key_values: Dict[TKey, Tuple[torch.Tensor, np.ndarray]] = {}
    all_annotator_change_device_times = []
    for bundle, detect_resolution, extra_annotator_params in zip(
        controls, detect_resolutions, extra_annotator_params_list
    ):
        i_type = bundle.type
        i_data = bundle.data
        i_hint_image = image_array_d[get_hint_url_key(i_data.hint_url)]
        i_t_annotator = i_data.hint_annotator or i_type
        api.prepare_annotator(i_t_annotator)
        i_annotator = api.annotators.get(i_t_annotator)
        i_bypass_annotator = i_data.bypass_annotator or i_annotator is None
        key = get_bundle_key(bundle)
        all_keys.append(key)
        i_value = all_key_values.get(key)
        if i_value is not None:
            continue
        device = api.device
        use_half = api.use_half
        i_binarize = i_data.hint_binarize_threshold is not None
        if i_bypass_annotator:
            i_o_hint_arr = i_hint_image
        else:
            if detect_resolution is not None:
                i_hint_image = resize_image(i_hint_image, detect_resolution)
            ht = time.time()
            if need_change_device:
                device = "cuda:0"
                use_half = True
                i_annotator.to(device, use_half=True)
            all_annotator_change_device_times.append(time.time() - ht)
            i_hint_kw = i_data.dict()
            if extra_annotator_params is not None:
                i_hint_kw.update(extra_annotator_params)
            i_o_hint_arr = api.get_hint_of(i_t_annotator, i_hint_image, **i_hint_kw)
            if i_binarize:
                binarized = i_o_hint_arr > i_data.hint_binarize_threshold
                i_o_hint_arr = np.where(binarized, 255, 0).astype(np.uint8)
            ht = time.time()
            if need_change_device:
                i_annotator.to("cpu", use_half=False)
                torch.cuda.empty_cache()
            all_annotator_change_device_times.append(time.time() - ht)
        interpolation = cv2.INTER_NEAREST if i_binarize else cv2.INTER_LINEAR
        i_hint_array = cv2.resize(i_o_hint_arr, (w, h), interpolation=interpolation)
        i_hint = torch.from_numpy(i_hint_array)[None].permute(0, 3, 1, 2)
        if use_half:
            i_hint = i_hint.half()
        i_hint = i_hint.contiguous().to(device) / 255.0
        all_key_values[key] = i_hint, i_o_hint_arr, i_binarize
    change_annotator_device_time = sum(all_annotator_change_device_times)
    t2 = time.time()
    # gather scales
    num_scales = api.m.num_control_scales
    all_scales = []
    for bundle in controls:
        i_data = bundle.data
        all_scales.append(
            [
                i_data.control_strength * (0.825 ** float(12 - i))
                for i in range(num_scales)
            ]
            if i_data.guess_mode
            else ([i_data.control_strength] * num_scales)
        )
    api.m.control_scales = all_scales
    cond = [common.prompt] * common.num_samples
    kw = shallow_copy_dict(kwargs)
    kw.update(handle_diffusion_model(api, common))
    hint = [(b.type, all_key_values[k][0]) for b, k in zip(controls, all_keys)]
    kw[CONTROL_HINT_KEY] = hint
    kw[CONTROL_HINT_START_KEY] = [b.data.hint_start for b in controls]
    kw[CONTROL_HINT_END_KEY] = [b.data.hint_end for b in controls]
    kw["max_wh"] = common.max_wh
    dt = time.time()
    if need_change_device:
        api.to("cuda:0", use_half=True, no_annotator=True)
    change_diffusion_device_time = time.time() - dt
    # inpainting workaround
    common = common.copy()
    is_sd_inpainting = common.use_inpainting
    common.use_raw_inpainting = not is_sd_inpainting
    if common.mask_url is not None or is_sd_inpainting:
        if common.url is None:
            raise ValueError("`url` should be provided to inpainting")
        if common.mask_url is None:
            raise ValueError("`mask_url` should be provided to inpainting")
        inpainting_mask = Image.fromarray(image_array_d[common.mask_url])
        image = Image.fromarray(image_array_d[common.url])
        if common.inpainting_target_wh is not None:
            it_wh = common.inpainting_target_wh
            if isinstance(it_wh, int):
                it_wh = it_wh, it_wh
            it_w, it_h = restrict_wh(*it_wh, common.max_wh)
            it_w = get_suitable_size(it_w, 64)
            it_h = get_suitable_size(it_h, 64)
            common.inpainting_target_wh = it_w, it_h
        kw.update(handle_diffusion_inpainting_model(common))
        image = image.resize((w, h))
        inpainting_mask = inpainting_mask.resize((w, h))
        outs = api.txt2img_inpainting(cond, image, inpainting_mask, **kw)
    elif common.url is None:
        kw["size"] = w, h
        outs = api.txt2img(cond, **kw)
    else:
        init_image = cv2.resize(
            image_array_d[common.url],
            (w, h),
            interpolation=cv2.INTER_LINEAR,
        )
        init_image = init_image[None].repeat(common.num_samples, axis=0)
        init_image = init_image.transpose([0, 3, 1, 2])
        init_image = init_image.astype(np.float32) / 255.0
        kw["cond"] = cond
        kw["fidelity"] = common.fidelity
        init_tensor = torch.from_numpy(init_image)
        if api.use_half:
            init_tensor = init_tensor.half()
        init_tensor = init_tensor.to(api.device)
        outs = api.img2img(init_tensor, **kw)
    dt = time.time()
    change_diffusion_device_time += time.time() - dt
    outs = 0.5 * (outs + 1.0)
    outs = to_uint8(outs).permute(0, 2, 3, 1).cpu().numpy()
    t3 = time.time()
    results = list(
        map(
            resize_to_original,
            *zip(*[p[1:] for p in all_key_values.values()]),
        )
    )
    for i in range(common.num_samples):
        results.append(resize_to_original(outs[i], False))
    latencies = dict(
        get_model=t_sd2 - t_sd,
        download_images=t0 - t_sd2,
        switch_control=t1 - t0,
        get_hint=t2 - t1 - change_annotator_device_time,
        change_annotator_device=change_annotator_device_time,
        inference=t3 - t2 - change_diffusion_device_time,
        change_diffusion_device=change_diffusion_device_time,
        post_resize=time.time() - t3,
    )
    return results, latencies


async def run_single_control(
    self: IAlgorithm,
    data: ControlNetModel,
    hint_type: ControlNetHints,
    **kwargs: Any,
) -> Tuple[List[np.ndarray], Dict[str, float]]:
    self.log_endpoint(data)
    bundle = ControlNetBundle(type=hint_type, data=data)
    return await apply_control(self, data, [bundle], **kwargs)


def register_control(
    algorithm_model_class: Type[ControlNetModel],
    algorithm_endpoint: str,
    hint_type: ControlNetHints,
) -> None:
    class _(IAlgorithm):
        model_class = algorithm_model_class

        endpoint = algorithm_endpoint

        def initialize(self) -> None:
            register_sd()
            register_sd_inpainting()

        async def run(
            self, data: algorithm_model_class, *args: Any, **kwargs: Any
        ) -> Response:
            rs, latencies = await run_single_control(self, data, hint_type, **kwargs)
            t0 = time.time()
            content = None if data.return_arrays else np_to_bytes(to_canvas(rs))
            t1 = time.time()
            latencies["to_canvas"] = t1 - t0
            self.log_times(latencies)
            if data.return_arrays:
                return rs
            return Response(content=content, media_type="image/png")

    IAlgorithm.auto_register()(_)
    control_hint2endpoints[hint_type] = algorithm_endpoint
    control_hint2data_models[hint_type] = algorithm_model_class


def register_hint(
    hint_model_class: Type,
    hint_endpoint: str,
    hint_type: ControlNetHints,
) -> None:
    class _(IAlgorithm):
        model_class = hint_model_class
        model_class.__name__ = hint_model_class.__name__

        endpoint = hint_endpoint

        def initialize(self) -> None:
            register_sd()

        async def run(
            self, data: hint_model_class, *args: Any, **kwargs: Any
        ) -> Response:
            t0 = time.time()
            image = await self.get_image_from("url", data, kwargs)
            w, h = image.size
            t1 = time.time()
            m = api_pool.get(APIs.SD)
            t2 = time.time()
            image = to_contrast_rgb(image)
            hint_image = np.array(image)
            detect_resolution = getattr(data, "detect_resolution", None)
            if detect_resolution is not None:
                hint_image = resize_image(hint_image, detect_resolution)
            hint = m.get_hint_of(hint_type, hint_image, **data.dict())
            hint = cv2.resize(hint, (w, h), interpolation=cv2.INTER_LINEAR)
            t3 = time.time()
            self.log_times(
                dict(
                    download=t1 - t0,
                    get_model=t2 - t1,
                    inference=t3 - t2,
                    cleanup=time.time() - t3,
                )
            )
            if data.return_arrays:
                return [hint]
            return Response(content=np_to_bytes(hint), media_type="image/png")

    IAlgorithm.auto_register()(_)
    control_hint2hint_endpoints[hint_type] = hint_endpoint
    control_hint2hint_data_models[hint_type] = hint_model_class


class DetectResolutionModel(BaseModel):
    detect_resolution: int = Field(
        384,
        ge=128,
        le=1024,
        description="Detect resolution.",
    )


class LargeDetectResolutionModel(BaseModel):
    detect_resolution: int = Field(
        512,
        ge=128,
        le=1024,
        description="Detect resolution.",
    )


# ControlNet (depth2image)


class _DepthModel(DetectResolutionModel):
    pass


class ControlDepthModel(_DepthModel, ControlNetModel):
    pass


class ControlDepthHintModel(_DepthModel, ReturnArraysModel, ImageModel):
    pass


# ControlNet (canny2image)


class _CannyModel(LargeDetectResolutionModel):
    low_threshold: int = Field(
        100,
        ge=1,
        le=255,
        description="Low threshold of canny algorithm.",
    )
    high_threshold: int = Field(
        200,
        ge=1,
        le=255,
        description="High threshold of canny algorithm.",
    )


class ControlCannyModel(_CannyModel, ControlNetModel):
    pass


class ControlCannyHintModel(_CannyModel, ReturnArraysModel, ImageModel):
    pass


# ControlNet (pose2image)


class _PoseModel(LargeDetectResolutionModel):
    pass


class ControlPoseModel(_PoseModel, ControlNetModel):
    pass


class ControlPoseHintModel(_PoseModel, ReturnArraysModel, ImageModel):
    pass


# ControlNet (mlsd2image)


class _MLSDModel(LargeDetectResolutionModel):
    value_threshold: float = Field(
        0.1,
        ge=0.01,
        le=2.0,
        description="Value threshold of mlsd model.",
    )
    distance_threshold: float = Field(
        0.1,
        ge=0.01,
        le=20.0,
        description="Distance threshold of mlsd model.",
    )


class ControlMLSDModel(_MLSDModel, ControlNetModel):
    pass


class ControlMLSDHintModel(_MLSDModel, ReturnArraysModel, ImageModel):
    pass


# register

control_hint2endpoints: Dict[ControlNetHints, str] = {}
control_hint2hint_endpoints: Dict[ControlNetHints, str] = {}
control_hint2data_models: Dict[ControlNetHints, Type[BaseModel]] = {}
control_hint2hint_data_models: Dict[ControlNetHints, Type[BaseModel]] = {}
register_control(ControlDepthModel, new_control_depth_endpoint, ControlNetHints.DEPTH)
register_control(ControlCannyModel, new_control_canny_endpoint, ControlNetHints.CANNY)
register_control(ControlPoseModel, new_control_pose_endpoint, ControlNetHints.POSE)
register_control(ControlMLSDModel, new_control_mlsd_endpoint, ControlNetHints.MLSD)
register_hint(ControlDepthHintModel, control_depth_hint_endpoint, ControlNetHints.DEPTH)
register_hint(ControlCannyHintModel, control_canny_hint_endpoint, ControlNetHints.CANNY)
register_hint(ControlPoseHintModel, control_pose_hint_endpoint, ControlNetHints.POSE)
register_hint(ControlMLSDHintModel, control_mlsd_hint_endpoint, ControlNetHints.MLSD)


__all__ = [
    "get_bundle_key",
    "new_control_depth_endpoint",
    "new_control_canny_endpoint",
    "new_control_pose_endpoint",
    "new_control_mlsd_endpoint",
    "control_depth_hint_endpoint",
    "control_canny_hint_endpoint",
    "control_pose_hint_endpoint",
    "control_mlsd_hint_endpoint",
    "ControlDepthModel",
    "ControlDepthHintModel",
    "ControlCannyModel",
    "ControlCannyHintModel",
    "ControlPoseModel",
    "ControlPoseHintModel",
    "ControlMLSDModel",
    "ControlMLSDHintModel",
    "control_hint2endpoints",
    "control_hint2hint_endpoints",
    "control_hint2data_models",
    "control_hint2hint_data_models",
]



import os
import cv2
import time
import torch

import numpy as np

from enum import Enum
from typing import Any
from typing import List
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cflearn.parameters import OPT
from cflearn.misc.toolkit import eval_context
from torchvision.transforms.functional import normalize

from ..cv import CVImageModel
from ..common import get_response
from ..common import IAlgorithm

try:
    from facexlib.parsing import init_parsing_model
    from facexlib.detection import init_detection_model
except:
    init_parsing_model = None
    init_detection_model = None


def get_root() -> str:
    return os.path.join(OPT.external_dir, "facexlib")


facexlib_parse_endpoint = "/third_party/facexlib/parse"
facexlib_detect_endpoint = "/third_party/facexlib/detect"


class FaceAreas(str, Enum):
    FACE = "face"
    NECK = "neck"
    HAIR = "hair"
    HAT = "hat"


class FacexlibParseModel(CVImageModel):
    mask_size: int = Field(0, description="mask size")
    affected_areas: List[FaceAreas] = Field(
        default_factory=lambda: [FaceAreas.FACE],
        description="affected face areas",
    )


@IAlgorithm.auto_register()
class FacexlibParse(IAlgorithm):
    model_class = FacexlibParseModel

    endpoint = facexlib_parse_endpoint

    def initialize(self) -> None:
        if init_parsing_model is None:
            raise ImportError("`facexlib` should be installed for `FacexlibParse`")
        self.model = init_parsing_model(device="cpu", model_rootpath=get_root())

    async def run(
        self, data: FacexlibParseModel, *args: Any, **kwargs: Any
    ) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        model = self.model.to("cuda")
        t2 = time.time()
        w, h = image.size
        array = np.array(image)
        if w != 512 or h != 512:
            rw = (int(w * (512 / w)) // 8) * 8
            rh = (int(h * (512 / h)) // 8) * 8
            array = cv2.resize(array, dsize=(rw, rh))
        array = array.astype(np.float32) / 255.0
        tensor = torch.from_numpy(array.transpose(2, 0, 1))
        normalize(tensor, (0.5, 0.5, 0.5), (0.5, 0.5, 0.5), inplace=True)
        tensor = torch.unsqueeze(tensor, 0).to("cuda")
        with eval_context(model):
            face = model(tensor)[0]
        t3 = time.time()
        face = face.squeeze(0).cpu().numpy().argmax(0)
        face = face.copy().astype(np.uint8)
        mask = self._to_mask(face, data.affected_areas)
        if data.mask_size > 0:
            mask = cv2.dilate(
                mask,
                np.ones((5, 5), np.uint8),
                iterations=data.mask_size,
            )
        if w != 512 or h != 512:
            mask = cv2.resize(mask, dsize=(w, h))
        t4 = time.time()
        self.model.to("cpu")
        t5 = time.time()
        res = get_response(data, [mask])
        self.log_times(
            {
                "download": t1 - t0,
                "to_cuda": t2 - t1,
                "inference": t3 - t2,
                "postprocess": t4 - t3,
                "cleanup": t5 - t4,
                "get_response": time.time() - t5,
            }
        )
        return res

    def _to_mask(self, face: np.ndarray, affected_areas: List[FaceAreas]) -> np.ndarray:
        area_set = set(affected_areas)
        keep_face = FaceAreas.FACE in area_set
        keep_neck = FaceAreas.NECK in area_set
        keep_hair = FaceAreas.HAIR in area_set
        keep_hat = FaceAreas.HAT in area_set

        mask = np.zeros((face.shape[0], face.shape[1], 3), dtype=np.uint8)
        num_of_class = np.max(face)
        for i in range(1, num_of_class + 1):
            index = np.where(face == i)
            if i < 14 and keep_face:
                mask[index[0], index[1], :] = [255, 255, 255]
            elif i == 14 and keep_neck:
                mask[index[0], index[1], :] = [255, 255, 255]
            elif i == 17 and keep_hair:
                mask[index[0], index[1], :] = [255, 255, 255]
            elif i == 18 and keep_hat:
                mask[index[0], index[1], :] = [255, 255, 255]
        return mask


class FacexlibDetectModel(CVImageModel):
    confidence: float = Field(0.97, description="confidence threshold")


class FacexlibDetectResponse(BaseModel):
    ltrbs: List[List[int]]


@IAlgorithm.auto_register()
class FacexlibDetect(IAlgorithm):
    model_class = FacexlibDetectModel
    response_model_class = FacexlibDetectResponse

    endpoint = facexlib_detect_endpoint

    def initialize(self) -> None:
        if init_detection_model is None:
            raise ImportError("`facexlib` should be installed for `FacexlibDetect`")
        self.model = init_detection_model(
            "retinaface_resnet50",
            device="cuda",
            model_rootpath=get_root(),
        )
        self.model.to("cpu")

    async def run(
        self, data: FacexlibDetectModel, *args: Any, **kwargs: Any
    ) -> FacexlibDetectResponse:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        model = self.model.to("cuda")
        t2 = time.time()
        with eval_context(model):
            boxes, _ = model.align_multi(image, data.confidence)
        t3 = time.time()
        ltrbs = [list(map(int, box[:4])) for box in boxes]
        t4 = time.time()
        self.model.to("cpu")
        self.log_times(
            {
                "download": t1 - t0,
                "to_cuda": t2 - t1,
                "inference": t3 - t2,
                "postprocess": t4 - t3,
                "cleanup": time.time() - t4,
            }
        )
        return FacexlibDetectResponse(ltrbs=ltrbs)


__all__ = [
    "facexlib_parse_endpoint",
    "facexlib_detect_endpoint",
    "FaceAreas",
    "FacexlibParseModel",
    "FacexlibDetectModel",
    "FacexlibDetectResponse",
    "FacexlibParse",
    "FacexlibDetect",
]



from typing import Dict
from pydantic import Field
from pydantic import BaseModel
from cfclient.models import ImageModel
from cflearn.api.cv import SDVersions

from ..common import MaxWHModel
from ..common import UseAuditModel
from ..common import DiffusionModel
from ..common import ReturnArraysModel


class _ControlNetModel(UseAuditModel):
    hint_url: str = Field(
        "",
        description="""
The `cdn` / `cos` url of the user's hint image.
> If empty string is provided, we will use `url` as `hint_url`.
> `cos` url from `qcloud` is preferred.
""",
    )
    hint_starts: Dict[str, float] = Field(
        default_factory=lambda: {},
        description="start ratio of each hint",
    )
    hint_ends: Dict[str, float] = Field(
        default_factory=lambda: {},
        description="end ratio of each hint",
    )
    prompt: str = Field(..., description="Prompt.")
    fidelity: float = Field(
        0.05,
        ge=0.0,
        le=1.0,
        description="The fidelity of the input image, only take effects when `use_img2img` is True.",
    )
    use_img2img: bool = Field(True, description="Whether use img2img method.")
    num_samples: int = Field(1, ge=1, le=4, description="Number of samples.")
    bypass_annotator: bool = Field(False, description="Bypass the annotator.")
    base_model: str = Field(
        SDVersions.v1_5,
        description="The base model.",
    )
    guess_mode: bool = Field(False, description="Guess mode.")
    no_switch: bool = Field(
        False,
        description="Whether not to switch the ControlNet weights even when the base model has switched.",
    )

    @property
    def version(self) -> str:
        return self.base_model


# only useful when inpainting model is used
class _InpaintingMixin(BaseModel):
    use_latent_guidance: bool = Field(
        False,
        description="Whether use the latent of the givent image to guide the generation.",
    )
    reference_fidelity: float = Field(
        0.0, description="Fidelity of the reference image."
    )


class ControlNetModel(
    ReturnArraysModel,
    _InpaintingMixin,
    DiffusionModel,
    MaxWHModel,
    _ControlNetModel,
    ImageModel,
):
    pass



import os
import json
import requests

from cfcreator import *
from cfcreator.sdks import *
from io import BytesIO
from PIL import Image
from tqdm import tqdm
from cftool.misc import print_info


test_parameters = {
    txt2img_sd_endpoint: {"text": "A lovely little cat."},
    txt2img_sd_inpainting_endpoint: {
        "text": "a large tree",
        "url": "https://ailabcdn.nolibox.com/tmp/3cf1b7cc46fc496bb758ed56c20f8195.png",
        "mask_url": "https://ailabcdn.nolibox.com/tmp/31c8575a2fd24a9d9a840266073a0523.png",
    },
    txt2img_sd_outpainting_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/upload/images/f1b0e606c8df49efa0d6dbf831576997.png",
        "text": "Blue cars are driving on the road, lakes, sunset, clean sky, superb",
    },
    txt2txt_prompt_enhance_endpoint: {"text": "A lovely little cat."},
    img2img_sd_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/upload/images/100cdc5685c3446eb1d7ce57e70eb682.png",
        "text": "(((masterpiece))),(((best quality))),((ultra-detailed)),((an extremely delicate and beautiful)), (beautiful detailed eyes), 1girl, (undine), (touhou), (((((komeiji_koishi)))))",
        "version": "anime_anything",
    },
    img2img_semantic2img_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/upload/images/1137f95fed6a44ce85d176131494a407.png",
        "color2label": {
            "rgb(170,170,170)": 106,
            "rgb(29,195,49)": 124,
            "rgb(200,146,117)": 127,
            "rgb(116,77,52)": 135,
            "rgb(157,157,255)": 148,
            "rgb(54,62,167)": 155,
            "rgb(95,219,255)": 157,
            "rgb(230,182,19)": 162,
            "rgb(128,255,255)": 178,
        },
    },
    img2img_sr_endpoint: {
        "url": "https://ailabcdn.nolibox.com/tmp/3cf1b7cc46fc496bb758ed56c20f8195.png",
        "is_anime": True,
        "target_w": 1024,
        "target_h": 1024,
    },
    img2img_inpainting_endpoint: {
        "model": "lama",
        "url": "https://ailab-huawei-cdn.nolibox.com/upload/images/ff9b03566cc6448498c1845ee12fc8b8.png",
        "mask_url": "https://ailab-huawei-cdn.nolibox.com/upload/images/bdf4cf43e4974922917fd5a93c4f335c.png",
    },
    img2img_harmonization_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/upload/images/ff9b03566cc6448498c1845ee12fc8b8.png",
        "mask_url": "https://ailab-huawei-cdn.nolibox.com/upload/images/bdf4cf43e4974922917fd5a93c4f335c.png",
        "strength": 1.0,
    },
    img2img_sod_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/upload/images/ff9b03566cc6448498c1845ee12fc8b8.png"
    },
    img2txt_caption_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/upload/images/ff9b03566cc6448498c1845ee12fc8b8.png"
    },
    control_depth_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png",
        "prompt": "a lovely little cat",
        "negative_prompt": "appendix, malformed, disfigured, ugly, deformed, mutilated, morbid, bad anatomy, cropped, blurred, mutated, error, lowres, blurry, low quality, signature, watermark, text",
        "fidelity": 0.05,
        "use_img2img": True,
        "seed": 123,
    },
    control_canny_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png",
        "prompt": "a lovely little cat",
        "negative_prompt": "appendix, malformed, disfigured, ugly, deformed, mutilated, morbid, bad anatomy, cropped, blurred, mutated, error, lowres, blurry, low quality, signature, watermark, text",
        "fidelity": 0.05,
        "use_img2img": True,
        "seed": 123,
    },
    control_pose_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png",
        "prompt": "a lovely little cat",
        "negative_prompt": "appendix, malformed, disfigured, ugly, deformed, mutilated, morbid, bad anatomy, cropped, blurred, mutated, error, lowres, blurry, low quality, signature, watermark, text",
        "fidelity": 0.05,
        "use_img2img": True,
        "seed": 123,
    },
    control_mlsd_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png",
        "prompt": "a lovely little cat",
        "negative_prompt": "appendix, malformed, disfigured, ugly, deformed, mutilated, morbid, bad anatomy, cropped, blurred, mutated, error, lowres, blurry, low quality, signature, watermark, text",
        "fidelity": 0.05,
        "use_img2img": True,
        "seed": 123,
    },
    control_multi_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png",
        "types": ["canny", "depth"],
        "prompt": "a lovely little cat",
        "negative_prompt": "appendix, malformed, disfigured, ugly, deformed, mutilated, morbid, bad anatomy, cropped, blurred, mutated, error, lowres, blurry, low quality, signature, watermark, text",
        "fidelity": 0.05,
        "use_img2img": True,
        "seed": 123,
    },
    control_depth_hint_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png"
    },
    control_canny_hint_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png"
    },
    control_pose_hint_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png"
    },
    control_mlsd_hint_endpoint: {
        "url": "https://ailab-huawei-cdn.nolibox.com/aigc-private/images/ef470a0e9566412aabdd17284db55675.png"
    },
}


def download_image(url: str) -> Image.Image:
    return Image.open(BytesIO(requests.get(url).content))


async def main(host: str):
    results = {}
    exceptions = {}
    image_folder = "images"
    for endpoint, params in tqdm(list(test_parameters.items())):
        uid = await push(host, endpoint, params)
        res = await poll(host, uid)
        task = endpoint2algorithm(endpoint)
        status = res["status"]
        (results if status == Status.FINISHED else exceptions)[task] = res
        if task.startswith("control"):
            task_folder = os.path.join(image_folder, task)
            os.makedirs(task_folder, exist_ok=True)
            response = res["data"]["response"]
            for i, hint_url in enumerate(response["hint_urls"]):
                i_path = os.path.join(task_folder, f"hint_{i}.png")
                download_image(hint_url).save(i_path)
            for i, result_url in enumerate(response["result_urls"]):
                i_path = os.path.join(task_folder, f"result_{i}.png")
                download_image(result_url).save(i_path)
            continue
        cdn = res.get("data", {}).get("cdn")
        if cdn is not None:
            os.makedirs(image_folder, exist_ok=True)
            image_path = os.path.join(image_folder, f"{task}.png")
            download_image(cdn).save(image_path)
    with open("results.json", "w") as f:
        json.dump(results, f, indent=2)
    with open("exceptions.json", "w") as f:
        json.dump(exceptions, f, indent=2)
    print_info(f"num success   : {len(results)}")
    print_info(f"num exception : {len(exceptions)}")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main("https://your-kafka-host.com"))



# `kafka` sdk is used to access to the server launched by `uvicorn apis.kafka.producer:app --host 0.0.0.0 --port 8123`

import json
import time
import requests

from typing import Any
from typing import Dict
from typing import Callable
from typing import Optional
from cftool.misc import get_err_msg
from cfclient.utils import get
from cfclient.utils import post
from cfcreator.common import endpoint2algorithm
from cfcreator.common import Status

from .utils import *


TCallback = Callable[[Dict[str, Any]], None]


async def push(host: str, endpoint: str, params: Dict[str, Any]) -> str:
    url = f"{host}/push"
    data = dict(task=endpoint2algorithm(endpoint), params=params)
    async with get_http_session() as session:
        while True:
            try:
                res = await post(url, data, session)
                return res["uid"]
            except Exception as err:
                print(f"error occurred when pushing: {get_err_msg(err)}")
                time.sleep(1)


def push_sync(host: str, endpoint: str, params: Dict[str, Any]) -> str:
    url = f"{host}/push"
    data = dict(task=endpoint2algorithm(endpoint), params=params)
    return requests.post(url, json=data).json()["uid"]


async def poll(
    host: str,
    uid: str,
    *,
    callback: Optional[TCallback] = None,
) -> Dict[str, Any]:
    url = f"{host}/status/{uid}"
    async with get_http_session() as session:
        while True:
            try:
                res = json.loads(await get(url, session))
            except Exception as err:
                print(f"error occurred when polling: {get_err_msg(err)}")
                time.sleep(1)
                continue
            if callback is not None:
                callback(res)
            if res["status"] == Status.PENDING:
                time.sleep(res["pending"])
            elif res["status"] == Status.WORKING:
                time.sleep(1)
            else:
                return res


def poll_sync(
    host: str,
    uid: str,
    *,
    callback: Optional[TCallback] = None,
) -> Dict[str, Any]:
    while True:
        res = requests.get(f"{host}/status/{uid}").json()
        if callback is not None:
            callback(res)
        if res["status"] == Status.PENDING:
            time.sleep(res["pending"])
        elif res["status"] == Status.WORKING:
            time.sleep(1)
        else:
            return res


__all__ = [
    "push",
    "poll",
    "push_sync",
    "poll_sync",
]



import time

import numpy as np

from PIL import Image
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from fastapi import Response
from pydantic import Field
from cftool.data_structures import WorkNode
from cftool.data_structures import Workflow

from .common import get_response
from .common import IWrapperAlgorithm
from .common import UseAuditModel
from .common import ReturnArraysModel


workflow_endpoint = "/workflow"
WORKFLOW_IS_EXCEPTION_KEY = "$is_exception"
WORKFLOW_TARGET_RESPONSE_KEY = "$target"


class WorkflowModel(ReturnArraysModel, UseAuditModel):
    nodes: List[WorkNode] = Field(..., description="The nodes in the workflow.")
    target: str = Field(..., description="The target node.")
    intermediate: Optional[List[str]] = Field(
        None,
        description="The intermediate nodes that you want to collect results from.",
    )
    caches: Optional[Dict[str, Any]] = Field(None, description="The preset caches.")
    return_if_exception: bool = Field(
        False,
        description="Whether to return the intermediate results anyway if any exception occurs.",
    )
    save_intermediate_to_private: bool = Field(
        True,
        description="Whether to save intermediate results to private buckets.",
    )


@IWrapperAlgorithm.auto_register()
class WorkflowAlgorithm(IWrapperAlgorithm):
    model_class = WorkflowModel

    endpoint = workflow_endpoint

    async def run(self, data: WorkflowModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        workflow = Workflow()
        for node in data.nodes:
            workflow.push(node)
        t1 = time.time()
        results = await self.apis.execute(
            workflow,
            data.target,
            data.caches,
            return_if_exception=data.return_if_exception,
        )
        t2 = time.time()
        # check exception message
        exception_message = results.get(self.exception_message_key)
        is_exception = exception_message is not None
        res = {WORKFLOW_IS_EXCEPTION_KEY: is_exception}
        if is_exception:
            res[WORKFLOW_TARGET_RESPONSE_KEY] = [exception_message]
        else:
            ## fetch target
            target_result = results[data.target]
            if not target_result or not isinstance(target_result[0], Image.Image):
                res[WORKFLOW_TARGET_RESPONSE_KEY] = target_result
            else:
                arrays = list(map(np.array, target_result))
                res[WORKFLOW_TARGET_RESPONSE_KEY] = get_response(data, arrays)
        # fetch intermediate
        if data.intermediate is not None:
            if not data.return_arrays:
                msg = "`return_arrays` should be True when `intermediate` is specified."
                raise ValueError(msg)
            for key in data.intermediate:
                intermediate_result = results.get(key)
                if intermediate_result:
                    if isinstance(intermediate_result[0], Image.Image):
                        intermediate_result = list(map(np.array, intermediate_result))
                if intermediate_result is not None:
                    res[key] = intermediate_result
        latencies = {
            "get_workflow": t1 - t0,
            "inference": t2 - t1,
            "get_response": time.time() - t2,
        }
        self.log_times(latencies)
        self.last_latencies["inference_details"] = results[self.latencies_key]
        return res


__all__ = [
    "workflow_endpoint",
    "WORKFLOW_IS_EXCEPTION_KEY",
    "WORKFLOW_TARGET_RESPONSE_KEY",
    "WorkflowModel",
    "WorkflowAlgorithm",
]



import cv2
import math
import time
import torch
import asyncio

import numpy as np
import torchvision.transforms as T

from PIL import Image
from enum import Enum
from typing import Any
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from transformers import AutoModel
from transformers import AutoFeatureExtractor
from PIL.ImageFilter import GaussianBlur
from cftool.cv import to_rgb
from cftool.cv import to_uint8
from cftool.cv import ImageBox
from cftool.cv import ImageProcessor
from cftool.geometry import Point
from cftool.geometry import Matrix2D
from cftool.geometry import PivotType
from cftool.geometry import ExpandType
from cftool.geometry import Matrix2DProperties
from cfclient.models import ImageModel
from cflearn.misc.toolkit import eval_context

from .utils import get_contrast_bg
from .common import get_response
from .common import IAlgorithm
from .common import ReturnArraysModel


cv_blur_endpoint = "/cv/blur"
cv_grayscale_endpoint = "/cv/grayscale"
cv_erode_endpoint = "/cv/erode"
cv_resize_endpoint = "/cv/resize"
cv_affine_endpoint = "/cv/affine"
cv_get_mask_endpoint = "/cv/get_mask"
cv_inverse_endpoint = "/cv/inverse"
cv_fill_bg_endpoint = "/cv/fill_bg"
cv_get_size_endpoint = "/cv/get_size"
cv_modify_box_endpoint = "/cv/modify_box"
cv_generate_masks_endpoint = "/cv/generate_masks"
cv_crop_image_endpoint = "/cv/crop_image"
cv_histogram_match_endpoint = "/cv/hist_match"
cv_image_similarity_endpoint = "/cv/similarity"
cv_repositioning_endpoint = "/cv/repositioning"


class CVImageModel(ReturnArraysModel, ImageModel):
    pass


class BlurModel(CVImageModel):
    radius: int = Field(2, description="size of the kernel")


@IAlgorithm.auto_register()
class Blur(IAlgorithm):
    model_class = BlurModel

    endpoint = cv_blur_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: BlurModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        array = np.array(image.filter(GaussianBlur(data.radius)))
        t2 = time.time()
        res = get_response(data, [array])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


@IAlgorithm.auto_register()
class Grayscale(IAlgorithm):
    model_class = CVImageModel

    endpoint = cv_grayscale_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: CVImageModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        array = np.array(image.convert("L"))
        t2 = time.time()
        res = get_response(data, [array])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


def erode(array: np.ndarray, n_iter: int, kernel_size: int) -> np.ndarray:
    kernel = np.ones((kernel_size, kernel_size), np.uint8)
    return cv2.erode(array, kernel, iterations=n_iter)


class Resampling(str, Enum):
    NEAREST = "nearest"
    BOX = "box"
    BILINEAR = "bilinear"
    HAMMING = "hamming"
    BICUBIC = "bicubic"
    LANCZOS = "lanczos"


class ResamplingModel(BaseModel):
    resampling: Resampling = Field(Resampling.BILINEAR, description="resampling method")


def resize(image: Image.Image, w: int, h: int, resampling: Resampling) -> Image.Image:
    if resampling == Resampling.NEAREST:
        r = Image.NEAREST
    elif resampling == Resampling.BOX:
        r = Image.BOX
    elif resampling == Resampling.BILINEAR:
        r = Image.BILINEAR
    elif resampling == Resampling.HAMMING:
        r = Image.HAMMING
    elif resampling == Resampling.BICUBIC:
        r = Image.BICUBIC
    else:
        r = Image.LANCZOS
    return image.resize((w, h), r)


def affine(
    image: Image.Image,
    a: float,
    b: float,
    c: float,
    d: float,
    e: float,
    f: float,
    w: int,
    h: int,
    resampling: Resampling,
    wh_limit: int,
) -> np.ndarray:
    matrix2d = Matrix2D(a=a, b=b, c=c, d=d, e=e, f=f)
    properties = matrix2d.decompose()
    iw, ih = image.size
    nw = max(round(iw * abs(properties.w)), 1)
    nh = max(round(ih * abs(properties.h)), 1)
    if nw > wh_limit or nh > wh_limit:
        raise ValueError(f"image size ({nw}, {nh}) exceeds wh_limit ({wh_limit})")
    array = np.array(resize(image, nw, nh, resampling))
    properties.w = 1
    properties.h = 1 if properties.h > 0 else -1
    matrix2d = Matrix2D.from_properties(properties)
    return cv2.warpAffine(array, matrix2d.matrix, [w, h])


class ErodeModel(CVImageModel):
    n_iter: int = Field(1, description="number of iterations")
    kernel_size: int = Field(3, description="size of the kernel")
    threshold: int = Field(0, description="threshold of the alpha channel")
    padding: int = Field(8, description="padding for the image")


@IAlgorithm.auto_register()
class Erode(IAlgorithm):
    model_class = ErodeModel

    endpoint = cv_erode_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: ErodeModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        w, h = image.size
        array = np.array(image.convert("RGBA"))
        alpha = array[..., -1]
        padded = np.pad(alpha, (data.padding, data.padding), constant_values=0)
        binarized = cv2.threshold(padded, data.threshold, 255, cv2.THRESH_BINARY)[1]
        eroded = erode(binarized, data.n_iter, data.kernel_size)
        shrinked = eroded[data.padding : -data.padding, data.padding : -data.padding]
        merged_alpha = np.minimum(alpha, shrinked)
        array[..., -1] = merged_alpha
        rgb = array[..., :3].reshape([-1, 3])
        rgb[(merged_alpha == 0).ravel()] = 0
        array[..., :3] = rgb.reshape([h, w, 3])
        t2 = time.time()
        res = get_response(data, [array])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


class ResizeModel(ResamplingModel, CVImageModel):
    w: int = Field(..., description="width of the output image")
    h: int = Field(..., description="width of the output image")


@IAlgorithm.auto_register()
class Resize(IAlgorithm):
    model_class = ResizeModel

    endpoint = cv_resize_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: ResizeModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        resized = np.array(resize(image, data.w, data.h, data.resampling))
        t2 = time.time()
        res = get_response(data, [resized])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


class BaseAffineModel(BaseModel):
    a: float = Field(..., description="`a` of the affine matrix")
    b: float = Field(..., description="`b` of the affine matrix")
    c: float = Field(..., description="`c` of the affine matrix")
    d: float = Field(..., description="`d` of the affine matrix")
    e: float = Field(..., description="`e` of the affine matrix")
    f: float = Field(..., description="`f` of the affine matrix")
    wh_limit: int = Field(
        16384,
        description="maximum width or height of the output image",
    )


class AffineModel(ResizeModel, BaseAffineModel):
    pass


@IAlgorithm.auto_register()
class Affine(IAlgorithm):
    model_class = AffineModel

    endpoint = cv_affine_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: AffineModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        output = affine(
            image,
            data.a,
            data.b,
            data.c,
            data.d,
            data.e,
            data.f,
            data.w,
            data.h,
            data.resampling,
            data.wh_limit,
        )
        t2 = time.time()
        res = get_response(data, [output])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


class GetMaskModel(CVImageModel):
    get_inverse: bool = Field(False, description="Whether get the inverse mask.")
    binarize_threshold: Optional[int] = Field(
        None,
        description="If not None, will binarize the mask with this value.",
    )


@IAlgorithm.auto_register()
class GetMask(IAlgorithm):
    model_class = GetMaskModel

    endpoint = cv_get_mask_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: GetMaskModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        if image.mode == "RGBA":
            mask = np.array(image)[..., -1]
        else:
            mask = np.array(image.convert("L"))
        if data.get_inverse:
            mask = 255 - mask
        if data.binarize_threshold is not None:
            mask = np.where(mask > data.binarize_threshold, 255, 0)
        t2 = time.time()
        res = get_response(data, [mask])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


@IAlgorithm.auto_register()
class Inverse(IAlgorithm):
    model_class = CVImageModel

    endpoint = cv_inverse_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: CVImageModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        array = np.array(image)
        inversed = 255 - array
        t2 = time.time()
        res = get_response(data, [inversed])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


class FillBGModel(CVImageModel):
    bg: Optional[Union[int, Tuple[int, int, int]]] = Field(
        None,
        description="""
Target background color.
> If not specified, `get_contrast_bg` will be used to calculate the `bg`.
""",
    )


@IAlgorithm.auto_register()
class FillBG(IAlgorithm):
    model_class = FillBGModel

    endpoint = cv_fill_bg_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: FillBGModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        bg = data.bg
        if bg is None:
            bg = get_contrast_bg(image)
        if isinstance(bg, int):
            bg = bg, bg, bg
        image = to_rgb(image, bg)
        t2 = time.time()
        res = get_response(data, [np.array(image)])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


@IAlgorithm.auto_register()
class GetSize(IAlgorithm):
    model_class = ImageModel

    endpoint = cv_get_size_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: ImageModel, *args: Any, **kwargs: Any) -> List[int]:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        w, h = image.size
        self.log_times(
            {
                "download": t1 - t0,
                "process": time.time() - t1,
            }
        )
        return [w, h]


class LTRBModel(BaseModel):
    lt_rb: Tuple[int, int, int, int] = Field(
        ...,
        description="The left-top and right-bottom points.",
    )


class ModifyBoxModel(LTRBModel):
    w: Optional[int] = Field(None, description="The width of the image.")
    h: Optional[int] = Field(None, description="The height of the image.")
    padding: int = Field(0, description="The padding size.")
    to_square: bool = Field(False, description="Turn the box into a square box.")


@IAlgorithm.auto_register()
class ModifyBox(IAlgorithm):
    model_class = ModifyBoxModel

    endpoint = cv_modify_box_endpoint

    def initialize(self) -> None:
        pass

    async def run(
        self, data: ModifyBoxModel, *args: Any, **kwargs: Any
    ) -> List[List[int]]:
        self.log_endpoint(data)
        t0 = time.time()
        w = data.w
        h = data.h
        padding = data.padding
        box = ImageBox(*data.lt_rb)
        box = box.pad(padding, w=w, h=h)
        if data.to_square:
            box = box.to_square()
        self.log_times({"process": time.time() - t0})
        return [list(box.tuple)]


class GenerateMasksModel(ReturnArraysModel):
    w: int = Field(..., description="The width of the canvas.")
    h: int = Field(..., description="The height of the canvas.")
    lt_rb_list: List[Tuple[int, int, int, int]] = Field(..., description="The boxes.")
    merge: bool = Field(False, description="Whether merge the masks.")


@IAlgorithm.auto_register()
class GenerateMasks(IAlgorithm):
    model_class = GenerateMasksModel

    endpoint = cv_generate_masks_endpoint

    def initialize(self) -> None:
        pass

    async def run(
        self, data: GenerateMasksModel, *args: Any, **kwargs: Any
    ) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        canvas = np.zeros((data.h, data.w), dtype=np.uint8)
        results = []
        for l, t, r, b in data.lt_rb_list:
            i_canvas = canvas if data.merge else canvas.copy()
            i_canvas[t:b, l:r] = 255
            if not data.merge:
                results.append(i_canvas)
        if data.merge:
            results.append(canvas)
        t1 = time.time()
        res = get_response(data, results)
        self.log_times(
            {
                "process": t1 - t0,
                "get_response": time.time() - t1,
            }
        )
        return res


class CropImageModel(CVImageModel, LTRBModel):
    pass


@IAlgorithm.auto_register()
class CropImage(IAlgorithm):
    model_class = CropImageModel

    endpoint = cv_crop_image_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: CropImageModel, *args: Any, **kwargs: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        l, t, r, b = data.lt_rb
        image = image.crop((l, t, r, b))
        t2 = time.time()
        res = get_response(data, [np.array(image)])
        self.log_times(
            {
                "download": t1 - t0,
                "process": t2 - t1,
                "get_response": time.time() - t2,
            }
        )
        return res


class HistogramMatchModel(CVImageModel):
    bg_url: str = Field(..., description="The `cdn` / `cos` url of the background.")
    use_hsv: bool = Field(False, description="Whether use the HSV space to match.")
    strength: float = Field(1.0, description="Strength of the matching.")


@IAlgorithm.auto_register()
class HistogramMatch(IAlgorithm):
    model_class = HistogramMatchModel

    endpoint = cv_histogram_match_endpoint

    def initialize(self) -> None:
        pass

    async def run(
        self,
        data: HistogramMatchModel,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        bg = await self.get_image_from("bg_url", data, kwargs)
        t1 = time.time()
        to_normalized = lambda im: np.array(im).astype(np.float32) / 255.0
        rgba, bg_arr = map(to_normalized, [image, bg])
        rgb, alpha = rgba[..., :3], rgba[..., -1:]
        merged = rgb * alpha + bg_arr * (1.0 - alpha)
        merged, bg_arr = map(to_uint8, [merged, bg_arr])
        t2 = time.time()
        if data.use_hsv:
            merged = cv2.cvtColor(merged, cv2.COLOR_RGB2HSV)
            bg_arr = cv2.cvtColor(bg_arr, cv2.COLOR_RGB2HSV)
        adjusted = ImageProcessor.match_histograms(
            merged,
            bg_arr,
            alpha[..., 0] > 0,
            strength=data.strength,
        )
        if data.use_hsv:
            adjusted = cv2.cvtColor(adjusted, cv2.COLOR_HSV2RGB)
        t3 = time.time()
        res = get_response(data, [adjusted])
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "calculation": t3 - t2,
                "get_response": time.time() - t3,
            }
        )
        return res


class ImageSimilarityModel(ReturnArraysModel):
    url_0: Union[str, List[str]]
    url_1: Union[str, List[str]]
    batch_size: int = Field(4, description="batch size")
    skip_to_rgb: bool = False


class ImageSimilarityResponse(BaseModel):
    similarity: Union[float, List[List[float]]]


@IAlgorithm.auto_register()
class ImageSimilarity(IAlgorithm):
    model_class = ImageSimilarityModel
    response_model_class = ImageSimilarityResponse

    endpoint = cv_image_similarity_endpoint

    def initialize(self) -> None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model_ckpt = "nateraw/vit-base-beans"
        extractor = AutoFeatureExtractor.from_pretrained(model_ckpt)
        self.model = AutoModel.from_pretrained(model_ckpt).to(device)
        self.transform = T.Compose(
            [
                T.Resize(int((256 / 224) * extractor.size["height"])),
                T.CenterCrop(extractor.size["height"]),
                T.ToTensor(),
                T.Normalize(mean=extractor.image_mean, std=extractor.image_std),
            ]
        )

    async def run(
        self,
        data: ImageSimilarityModel,
        *args: Any,
        **kwargs: Any,
    ) -> ImageSimilarityResponse:
        self.log_endpoint(data)
        t0 = time.time()
        if isinstance(data.url_0, str) and isinstance(data.url_1, str):
            is_item = True
            futures = [
                self.get_image_from("url_0", data, kwargs),
                self.get_image_from("url_1", data, kwargs),
            ]
            im0_indices = [0]
            im1_indices = [1]
        else:
            is_item = False
            url_0 = data.url_0 if isinstance(data.url_0, list) else [data.url_0]
            url_1 = data.url_1 if isinstance(data.url_1, list) else [data.url_1]
            all_urls = {}
            im0_indices = []
            im1_indices = []
            for url in url_0:
                im0_indices.append(all_urls.setdefault(url, len(all_urls)))
            for url in url_1:
                im1_indices.append(all_urls.setdefault(url, len(all_urls)))
            reverse_urls = {v: k for k, v in all_urls.items()}
            sorted_urls = [reverse_urls[i] for i in range(len(all_urls))]
            futures = list(map(self.download_image_with_retry, sorted_urls))
        images = await asyncio.gather(*futures)
        t1 = time.time()
        if not data.skip_to_rgb:
            images = list(map(to_rgb, images))
        t2 = time.time()
        embeddings = self._extract_embeddings(images, data.batch_size)
        t3 = time.time()
        e0 = embeddings[im0_indices]
        e1 = embeddings[im1_indices]
        sim = (e0 @ e1.t()) / (e0.norm(dim=-1)[..., None] * e1.norm(dim=-1)[None])
        if is_item:
            sim = sim.item()
        else:
            sim = sim.cpu().numpy().tolist()
        self.log_times(
            {
                "download": t1 - t0,
                "to_rgb": t2 - t1,
                "inference": t3 - t2,
                "calculation": time.time() - t3,
            }
        )
        return ImageSimilarityResponse(similarity=sim)

    def _extract_embeddings(
        self,
        images: List[Image.Image],
        batch_size: int,
    ) -> torch.Tensor:
        transformed = torch.stack(list(map(self.transform, images)))
        with eval_context(self.model):
            num_batches = (len(images) - 1) // batch_size + 1
            embeddings = []
            for i in range(num_batches):
                i_transformed = transformed[i * batch_size : (i + 1) * batch_size]
                batch = {"pixel_values": i_transformed.to(self.model.device)}
                embeddings.append(self.model(**batch).last_hidden_state[:, 0].cpu())
        return torch.cat(embeddings)


class FillType(str, Enum):
    FIT = "fit"
    IOU = "iou"
    COVER = "cover"


class AlignType(str, Enum):
    CENTER = "center"
    CENTROID = "centroid"
    HALF_CENTROID = "half-centroid"


class AffineFrameModel(BaseModel):
    x: float
    y: float
    w: float
    h: float
    rotation: float
    frame_w: float
    frame_h: float

    @property
    def wh_ratio(self) -> float:
        return self.w / self.h

    @property
    def frame_wh_ratio(self) -> float:
        return self.frame_w / self.frame_h

    def scale_to(self, ow: float, oh: float) -> "Matrix2D":
        return Matrix2D.from_properties(
            Matrix2DProperties(
                x=self.x,
                y=self.y,
                w=self.w,
                h=self.h,
                theta=self.rotation * math.pi / 180,
                skew_x=0,
                skew_y=0,
            )
        ).scale(ow / self.frame_w, oh / self.frame_h, center=Point.origin())


class ScaleByDensityModel(BaseModel):
    min_scale: float = Field(
        0.8,
        description="The minimum scale used when density is large.",
    )
    max_scale: float = Field(
        1.1,
        description="The maximum scale used when density is small.",
    )


class RepositioningModel(ResamplingModel, CVImageModel):
    url: str = Field(..., description="The `cdn` / `cos` url of the goods.")
    output_width: int = Field(800, description="The width of the output image.")
    output_height: int = Field(800, description="The height of the output image.")
    binary_threshold: int = Field(127, description="The binary threshold.")
    stick_to_edge: bool = Field(
        True,
        description="Whether enable 'stick to edge' logic.",
    )
    stick_distance_threshold: int = Field(
        3,
        description="The distance threshold of 'sticking'.",
    )
    stick_image_fill_type: FillType = Field(
        FillType.FIT,
        description="The fill type under 'stick to edge' logic.",
    )
    stick_priorities: str = Field("tlbr", description="The priorities of 'sticking'.")
    affine_frame_list: Optional[List[AffineFrameModel]] = Field(
        None,
        description="The target affine frames to match.",
    )
    fill_type: FillType = Field(FillType.FIT, description="The fill type.")
    align_type: AlignType = Field(AlignType.CENTER, description="The align type.")
    scale_by_density: bool = Field(
        True,
        description="Whether enable 'scale by density' logic.",
    )
    scale_by_density_params: Optional[ScaleByDensityModel] = Field(
        None,
        description="The params of 'scale by density' logic.",
    )
    wh_limit: int = Field(
        16384,
        description="maximum width or height of the output image",
    )


opposite_edges = dict(l="r", t="b", r="l", b="t")


@IAlgorithm.auto_register()
class Repositioning(IAlgorithm):
    model_class = RepositioningModel

    endpoint = cv_repositioning_endpoint

    def initialize(self) -> None:
        pass

    async def run(
        self,
        data: RepositioningModel,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.get_image_from("url", data, kwargs)
        t1 = time.time()
        ow, oh = data.output_width, data.output_height
        image = image.convert("RGBA")
        array = np.array(image)
        alpha = array[..., -1]
        image_box = ImageBox.from_mask(alpha, data.binary_threshold)
        img_w, img_h = image.size
        sticky_edges = set()
        if data.stick_to_edge:
            if image_box.l < data.stick_distance_threshold:
                sticky_edges.add("l")
            if image_box.t < data.stick_distance_threshold:
                sticky_edges.add("t")
            if image_box.r > img_w - data.stick_distance_threshold:
                sticky_edges.add("r")
            if image_box.b > img_h - data.stick_distance_threshold:
                sticky_edges.add("b")
        t2 = time.time()
        mask = None
        mask_sum = None
        if sticky_edges:
            fill_type = data.stick_image_fill_type
            image_wh_ratio = img_w / img_h
            frame_matrix = Matrix2D(a=ow, b=0, c=0, d=oh, e=0, f=0)
            density_scale = None
        else:
            image = image.crop(image_box.tuple)
            img_w, img_h = image.size
            alpha = alpha[image_box.t : image_box.b, image_box.l : image_box.r]
            if not data.scale_by_density:
                density_scale = None
            else:
                density_scale_params = data.scale_by_density_params
                if density_scale_params is None:
                    density_scale_params = ScaleByDensityModel()
                min_density_scale = density_scale_params.min_scale
                max_density_scale = density_scale_params.max_scale
                mask = alpha > data.binary_threshold
                mask_sum = mask.sum()
                density = 1.0 - mask_sum / (max(image_box.w, image_box.h) ** 2)
                density_scale = (
                    min_density_scale
                    + (max_density_scale - min_density_scale) * density
                )
            fill_type = data.fill_type
            affine_frames = data.affine_frame_list
            if affine_frames is None:
                d = dict(x=40, y=40, w=720, h=720, rotation=0, frame_w=800, frame_h=800)
                affine_frames = [AffineFrameModel(**d)]
            canvas_wh_ratio = ow / oh
            filtered = [
                frame
                for frame in affine_frames
                if abs(frame.frame_wh_ratio - canvas_wh_ratio) < 1.0e-6
            ]
            if not filtered:
                raise ValueError("no affine frame matches the output ratio")
            wh_ratios = [frame.wh_ratio for frame in filtered]
            image_wh_ratio = image_box.wh_ratio
            best_index = np.argmin([abs(ratio - image_wh_ratio) for ratio in wh_ratios])
            frame_matrix = filtered[best_index].scale_to(ow, oh)
        frame_wh_ratio = frame_matrix.abs_wh_ratio
        if fill_type == FillType.IOU:
            expand_type = ExpandType.IOU
        elif fill_type == FillType.FIT:
            expand_type = (
                ExpandType.FIX_H
                if frame_wh_ratio > image_wh_ratio
                else ExpandType.FIX_W
            )
        else:
            expand_type = (
                ExpandType.FIX_H
                if frame_wh_ratio < image_wh_ratio
                else ExpandType.FIX_W
            )
        expanded_image_matrix = frame_matrix.set_wh_ratio(
            image_wh_ratio,
            type=expand_type,
            pivot=PivotType.CENTER,
        )
        if density_scale is not None:
            expanded_image_matrix = expanded_image_matrix.scale(
                density_scale,
                density_scale,
                center=expanded_image_matrix.pivot(PivotType.CENTER),
            )
        expanded_properties = expanded_image_matrix.decompose()
        ew, eh = expanded_properties.w, expanded_properties.h
        ew_ratio = ew / img_w
        eh_ratio = eh / img_h
        expanded_properties.w = ew_ratio
        expanded_properties.h = eh_ratio
        delta = None
        if sticky_edges:
            sticked = set()
            for edge in data.stick_priorities:
                if edge not in sticky_edges:
                    continue
                if opposite_edges[edge] in sticked:
                    continue
                if edge == "l":
                    expanded_properties.x = 0
                elif edge == "t":
                    expanded_properties.y = 0
                elif edge == "r":
                    expanded_properties.x = ow - ew
                elif edge == "b":
                    expanded_properties.y = oh - eh
                sticked.add(edge)
        elif data.align_type != AlignType.CENTER:
            if mask is None:
                mask = alpha > data.binary_threshold
            if mask_sum is None:
                mask_sum = mask.sum()
            x_center = 0.5 * img_w
            y_center = 0.5 * img_h
            x_centroid = (mask * np.arange(img_w)[None]).sum() / mask_sum
            y_centroid = (mask * np.arange(img_h)[..., None]).sum() / mask_sum
            x_delta = x_center - x_centroid.item()
            y_delta = y_center - y_centroid.item()
            if data.align_type == AlignType.HALF_CENTROID:
                x_delta *= 0.5
                y_delta *= 0.5
            delta = Point(x_delta, y_delta).rotate(frame_matrix.theta)
        final_matrix = Matrix2D.from_properties(expanded_properties)
        if delta is not None:
            final_matrix = final_matrix.move(delta)
        affined = affine(
            image,
            final_matrix.a,
            final_matrix.b,
            final_matrix.c,
            final_matrix.d,
            final_matrix.e,
            final_matrix.f,
            ow,
            oh,
            data.resampling,
            data.wh_limit,
        )
        t3 = time.time()
        res = get_response(data, [affined])
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "calculation": t3 - t2,
                "get_response": time.time() - t3,
            }
        )
        return res


__all__ = [
    "cv_blur_endpoint",
    "cv_grayscale_endpoint",
    "cv_erode_endpoint",
    "cv_resize_endpoint",
    "cv_affine_endpoint",
    "cv_get_mask_endpoint",
    "cv_inverse_endpoint",
    "cv_fill_bg_endpoint",
    "cv_get_size_endpoint",
    "cv_modify_box_endpoint",
    "cv_generate_masks_endpoint",
    "cv_crop_image_endpoint",
    "cv_histogram_match_endpoint",
    "cv_image_similarity_endpoint",
    "cv_repositioning_endpoint",
    "CVImageModel",
    "BlurModel",
    "ErodeModel",
    "ResizeModel",
    "BaseAffineModel",
    "AffineModel",
    "FillBGModel",
    "LTRBModel",
    "ModifyBoxModel",
    "GenerateMasksModel",
    "CropImageModel",
    "HistogramMatchModel",
    "ImageSimilarityModel",
    "ImageSimilarityResponse",
    "RepositioningModel",
    "Blur",
    "Grayscale",
    "Erode",
    "Resize",
    "Affine",
    "GetMask",
    "Inverse",
    "FillBG",
    "GetSize",
    "ModifyBox",
    "GenerateMasks",
    "CropImage",
    "HistogramMatch",
    "ImageSimilarity",
    "Repositioning",
]



import time

import numpy as np

from PIL import Image
from typing import Any
from typing import Dict
from typing import Tuple
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cftool.cv import to_rgb
from cftool.cv import to_uint8
from cfclient.models.core import ImageModel

from .cv import affine
from .cv import Resampling
from .cv import BaseAffineModel
from .cv import ResamplingModel
from .common import get_response
from .common import IAlgorithm
from .common import ReturnArraysModel


paste_pipeline_endpoint = "/pipeline/paste"


# paste pipeline


def paste(
    original_fg: Image.Image,
    original_bg: Image.Image,
    a: float,
    b: float,
    c: float,
    d: float,
    e: float,
    f: float,
    force_rgb: bool,
    resampling: Resampling,
    wh_limit: int,
) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
    t0 = time.time()
    if original_fg.mode != "RGBA":
        original_fg = original_fg.convert("RGBA")
    original_w, original_h = original_bg.size
    affined_fg_array = affine(
        original_fg,
        a,
        b,
        c,
        d,
        e,
        f,
        original_w,
        original_h,
        resampling,
        wh_limit,
    )
    t1 = time.time()
    affined_fg_array = affined_fg_array.astype(np.float32) / 255.0
    rgb = affined_fg_array[..., :3]
    mask = affined_fg_array[..., -1:]
    if force_rgb:
        original_bg = to_rgb(original_bg)
    bg_array = np.array(original_bg).astype(np.float32) / 255.0
    fg_array = rgb if bg_array.shape[2] == 3 else affined_fg_array
    merged = fg_array * mask + bg_array * (1.0 - mask)
    results = dict(rgb=rgb, mask=mask, merged=to_uint8(merged))
    latencies = {"affine": t1 - t0, "merge": time.time() - t1}
    return results, latencies


class _PastePipelineModel(BaseModel):
    bg_url: str = Field(
        ...,
        description="""
The `cdn` / `cos` url of the background's image.
> `cos` url from `qcloud` is preferred.
""",
    )
    force_rgb: bool = Field(False, description="Whether to force the output to be RGB.")
    return_mask: bool = Field(False, description="Whether to return the mask.")


class PastePipelineModel(
    ReturnArraysModel,
    ResamplingModel,
    BaseAffineModel,
    _PastePipelineModel,
    ImageModel,
):
    pass


@IAlgorithm.auto_register()
class PastePipeline(IAlgorithm):
    model_class = PastePipelineModel

    endpoint = paste_pipeline_endpoint

    def initialize(self) -> None:
        pass

    async def run(
        self,
        data: PastePipelineModel,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        original_fg = await self.get_image_from("url", data, kwargs)
        original_bg = await self.get_image_from("bg_url", data, kwargs)
        t1 = time.time()
        results, latencies = paste(
            original_fg,
            original_bg,
            data.a,
            data.b,
            data.c,
            data.d,
            data.e,
            data.f,
            data.force_rgb,
            data.resampling,
            data.wh_limit,
        )
        latencies["download"] = t1 - t0
        t2 = time.time()
        if not data.return_mask:
            res = get_response(data, [results["merged"]])
        else:
            mask = to_uint8(results["mask"])[..., 0]
            res = get_response(data, [results["merged"], mask])
        latencies["get_response"] = time.time() - t2
        self.log_times(latencies)
        return res


__all__ = [
    "paste_pipeline_endpoint",
    "PastePipelineModel",
    "PastePipeline",
]
